import kachery_client as kc
import hither2 as hi
import numpy as np
from sortingview.extractors.h5extractors.h5sortingextractorv1 import H5SortingExtractorV1, H5SortingExtractorV1Writer
from sortingview.serialize_wrapper import serialize_wrapper
import spikeextractors as se
from sortingview.config import job_cache, job_handler

from sortingview.extractors import LabboxEphysSortingExtractor
from ._correlograms_phy import compute_correlograms


@hi.function(
    'fetch_correlogram_plot_data', '0.2.7',
    image=hi.RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
@serialize_wrapper
def fetch_correlogram_plot_data(sorting_object, unit_x, unit_y=None):
    S = LabboxEphysSortingExtractor(sorting_object)
    data = _get_correlogram_data(sorting=S, unit_id1=unit_x, unit_id2=unit_y,
        window_size_msec=50, bin_size_msec=1)
    return data

def _get_time_segments(max_time: int, segment_size: int, num_segments: int):
    interval = np.floor(max_time / num_segments)
    start_times = [
        i * interval
        for i in range(num_segments)
    ]
    end_times = [
        i * interval + segment_size
        for i in range(num_segments)
    ]
    return start_times, end_times

@hi.function(
    'subsample_sorting', '0.1.5',
    image=hi.RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
def subsample_sorting(sorting_object: dict, segment_size_sec: float, num_segments: int):
    S = LabboxEphysSortingExtractor(sorting_object)
    segment_size = np.floor(S.get_sampling_frequency() * segment_size_sec)

    unit_ids = S.get_unit_ids()
    max_times = [np.max(S.get_unit_spike_train(unit_id)) for unit_id in unit_ids]
    max_time = np.floor(np.max(max_times))

    if max_time <= segment_size * num_segments * 2:
        return sorting_object
    
    segment_start_times, segment_end_times = _get_time_segments(max_time=max_time, segment_size=segment_size, num_segments=num_segments)
    with kc.TemporaryDirectory() as tmpdir:
        save_path = f'{tmpdir}/sorting.h5'
        W = H5SortingExtractorV1Writer(save_path=save_path, samplerate=S.get_sampling_frequency())
        for unit_id in unit_ids:
            st = S.get_unit_spike_train(unit_id)
            a = [st[(segment_start_times[i] <= st) & (st < segment_end_times[i])] for i in range(len(segment_start_times))]
            times_subsampled = np.concatenate(a)
            W.add_unit(unit_id=unit_id, times=times_subsampled)
        W.finalize()
        h5_path = kc.store_file(save_path)

    sorting_subsampled = LabboxEphysSortingExtractor({
        'sorting_format': 'h5_v1',
        'data': {
            'h5_path': h5_path
        }
    })
    return sorting_subsampled.object()

@kc.taskfunction('fetch_correlogram_plot_data.6', type='pure-calculation')
def task_fetch_correlogram_plot_data(*, sorting_object, unit_x, unit_y=None, subsample: bool):
    with hi.Config(
        job_cache=job_cache,
        job_handler=job_handler.correlograms
    ):
        with hi.Config(
            job_handler=job_handler.misc
        ):
            if subsample:
                sorting_object_subsampled = subsample_sorting.run(
                    sorting_object=sorting_object,
                    segment_size_sec=10,
                    num_segments=100
                )
            else:
                sorting_object_subsampled = sorting_object
        return fetch_correlogram_plot_data.run(
            sorting_object=sorting_object_subsampled,
            unit_x=unit_x,
            unit_y=unit_y
        )

def _get_spike_train(*, sorting: se.SortingExtractor, unit_id):
    if type(unit_id) == list:
        x = sorting.get_units_spike_train(unit_ids=unit_id)
        return np.sort(np.concatenate(x))
    else:
        return sorting.get_unit_spike_train(unit_id=unit_id)

def _get_correlogram_data(*, sorting, unit_id1, unit_id2=None, window_size_msec, bin_size_msec):
    auto = unit_id2 is None or unit_id2 == unit_id1

    times = _get_spike_train(sorting=sorting, unit_id=unit_id1)
    window_size = window_size_msec / 1000
    bin_size = bin_size_msec / 1000
    labels = np.ones(times.shape, dtype=np.int32)
    cluster_ids = [1]
    if not auto:
        times2 = _get_spike_train(sorting=sorting, unit_id=unit_id2)
        times = np.concatenate((times, times2))
        labels = np.concatenate((labels, np.ones(times2.shape, dtype=np.int32) *2 ))
        cluster_ids.append(2)
        inds = np.argsort(times)
        times = times[inds]
        labels = labels[inds]
    C = compute_correlograms(
        spike_times=times / sorting.get_sampling_frequency(),
        spike_clusters=labels,
        cluster_ids=cluster_ids,
        bin_size=bin_size,
        window_size=window_size,
        sample_rate=sorting.get_sampling_frequency(),
        symmetrize=True
    )
    bins = np.linspace(- window_size_msec / 2, window_size_msec / 2, C.shape[2])
    bin_counts = C[0, 0, :] if auto else C[0, 1, :]
    bin_size_sec = bin_size_msec / 1000
    return {
        'bins': bins.astype(np.float32),
        'bin_counts': bin_counts.astype(np.int32),
        'bin_size_sec': bin_size_sec
    }


import os

from pynwb import load_namespaces

# Set path of the namespace.yaml file to the expected install location
ndx_franklab_novela_specpath = os.path.join(
    os.path.dirname(__file__),
    'spec',
    'ndx-franklab-novela.namespace.yaml'
)

# If the extension has not been installed yet but we are running directly from
# the git repo
if not os.path.exists(ndx_franklab_novela_specpath):
    ndx_franklab_novela_specpath = os.path.abspath(os.path.join(
        os.path.dirname(__file__),
        '..', '..', '..',
        'spec',
        'ndx-franklab-novela.namespace.yaml'
    ))

# Load the namespace
load_namespaces(ndx_franklab_novela_specpath)



import math
from typing import Union
from sortingview.extractors import LabboxEphysRecordingExtractor, LabboxEphysSortingExtractor

class SpikeSortingView:
    def __init__(self, *,
        recording: LabboxEphysRecordingExtractor,
        sorting: LabboxEphysSortingExtractor,
        segment_duration_minutes: float
    ) -> None:
        self._recording = recording
        self._sorting = sorting
        self._segment_duration_minutes = segment_duration_minutes
    @property
    def recording(self):
        return self._recording
    @property
    def sorting(self):
        return self._sorting
    @property
    def segment_duration_minutes(self):
        return self._segment_duration_minutes
    @property
    def duration_minutes(self):
        return self._recording.get_num_frames() / self._recording.get_sampling_frequency() / 60
    @property
    def num_segments(self):
        return math.ceil(self.duration_minutes / self.segment_duration_minutes)
    @property
    def unit_ids(self):
        return self._sorting.get_unit_ids()
    @property
    def channel_ids(self):
        return self._recording.get_channel_ids()
    @property
    def sampling_frequency(self):
        return self._recording.get_sampling_frequency()
    def get_segment(self, start_index: int, end_index: Union[int, None]=None):
        from .SSVSegment import SSVSegment
        if end_index is None:
            end_index = start_index + 1
        start_frame = start_index * self.segment_duration_minutes * 60 * self.sampling_frequency
        end_frame = min(start_frame + self._segment_duration_minutes * 60 * self.sampling_frequency, self._recording.get_num_frames())
        return SSVSegment(parent=self, start_frame=start_frame, end_frame=end_frame)


import spikeextractors as se
import spikeinterface as si

class WrapperRecordingExtractor(se.RecordingExtractor):
    extractor_name = 'SpikeInterfaceRecordingExtractor'
    is_writable = False
    def __init__(self, new_recording: si.BaseRecording):
        se.RecordingExtractor.__init__(self)
        
        self._new_recording = new_recording

        for id in new_recording.get_channel_ids():
            pos = new_recording.get_channel_property(id, 'location')
            self.set_channel_property(id, 'location', pos)

    def get_channel_ids(self):
        return [int(id) for id in self._new_recording.channel_ids]

    def get_num_frames(self):
        return self._new_recording.get_num_frames()

    def get_sampling_frequency(self):
        return self._new_recording.get_sampling_frequency()

    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        X = self._new_recording.get_traces(
            segment_index=0,
            start_frame=start_frame,
            end_frame=end_frame,
            channel_ids=channel_ids,
        )
        return X.T


import math
from typing import List, Union, cast
import numpy as np
import h5py
import spikeextractors as se

class TimeseriesModel_Hdf5:
    def __init__(self, path):
        self._hdf5_path = path
        with h5py.File(self._hdf5_path, "r") as f:
            self._num_chunks = np.array(f.get('num_chunks'))[0]
            self._chunk_size = np.array(f.get('chunk_size'))[0]
            self._padding = np.array(f.get('padding'))[0]
            self._num_channels = np.array(f.get('num_channels'))[0]
            self._num_timepoints = np.array(f.get('num_timepoints'))[0]
            self._sampling_frequency = np.array(f.get('sampling_frequency'))[0]
            self._channel_ids = np.array(f.get('channel_ids')).astype(int).tolist()
            self._geom = np.array(f.get('channel_geom', None))

    def channelIds(self) -> List[int]:
        return self._channel_ids

    def numChannels(self):
        return int(self._num_channels)

    def numTimepoints(self):
        return int(self._num_timepoints)

    def samplingFrequency(self):
        return float(self._sampling_frequency)
    
    def geom(self) -> Union[np.ndarray, None]:
        return self._geom

    def getChunk(self, *, t1: int, t2: int, channel_inds: List[int]):
        if (t1 < 0) or (t2 > self.numTimepoints()):
            ret = np.zeros((len(channel_inds), t2-t1))
            t1a = max(t1, 0)
            t2a = min(t2, self.numTimepoints())
            ret[:, t1a-(t1):t2a-(t1)] = self.getChunk(t1=t1a, t2=t2a, channel_inds=channel_inds)
            return ret
        else:
            c1 = int(t1/self._chunk_size)
            c2 = int((t2-1)/self._chunk_size)
            ret = np.zeros((len(channel_inds), t2-t1))
            with h5py.File(self._hdf5_path, "r") as f:
                for cc in range(c1, c2+1):
                    if cc == c1:
                        t1a = t1
                    else:
                        t1a = self._chunk_size*cc
                    if cc == c2:
                        t2a = t2
                    else:
                        t2a = self._chunk_size*(cc+1)
                    for ii in range(len(channel_inds)):
                        m = channel_inds[ii]
                        assert(cc >= 0)
                        assert(cc < self._num_chunks)
                        str = 'part-{}-{}'.format(m, cc)
                        offset = self._chunk_size*cc-self._padding
                        ret[ii, t1a-t1:t2a-t1] = cast(np.ndarray, f[str])[t1a-offset:t2a-offset]
            return ret


class TimeseriesModel_Recording:
    def __init__(self, recording):
        self._recording = recording

    def numChannels(self):
        return len(self._recording.get_channel_ids())

    def numTimepoints(self):
        return self._recording.get_num_frames()

    def getChunk(self, *, t1, t2, channels):
        channel_ids = self._recording.get_channel_ids()
        channels2 = np.zeros(len(channels))
        for i in range(len(channels)):
            channels2[i] = channel_ids[int(channels[i])]
        if (t1 < 0) or (t2 > self.numTimepoints()):
            ret = np.zeros((len(channels), t2-t1))
            t1a = np.maximum(t1, 0)
            t2a = np.minimum(t2, self.numTimepoints())
            ret[:, t1a-(t1):t2a-(t1)] = self.getChunk(t1=t1a,
                                                      t2=t2a, channels=channels)
            return ret
        else:
            return self._recording.get_traces(start_frame=t1, end_frame=t2, channel_ids=channels2)

def set_geom_on_recording(recording: se.RecordingExtractor, geom: np.ndarray):
    channel_ids = recording.get_channel_ids()
    for ii in range(len(channel_ids)):
        recording.set_channel_property(channel_ids[ii], 'location', geom[ii, :].tolist())

def geom_from_recording(recording):
    channel_ids = recording.get_channel_ids()
    location0 = recording.get_channel_property(channel_ids[0], 'location')
    nd = len(location0)
    M = len(channel_ids)
    geom = np.zeros((M, nd))
    for ii in range(len(channel_ids)):
        location_ii = recording.get_channel_property(channel_ids[ii], 'location')
        geom[ii, :] = list(location_ii)
    return geom

def prepare_timeseries_hdf5_from_recording(recording: se.RecordingExtractor, timeseries_hdf5_fname: str, *, chunk_size: int, padding: int, dtype=float):
    chunk_size_with_padding = chunk_size+2*padding
    with h5py.File(timeseries_hdf5_fname, "w") as f:
        channel_ids = cast(List[int], recording.get_channel_ids())
        M = len(channel_ids)  # Number of channels
        N = cast(int, recording.get_num_frames())  # Number of timepoints
        num_chunks = math.ceil(N/chunk_size)
        f.create_dataset('chunk_size', data=[chunk_size])
        f.create_dataset('num_chunks', data=[num_chunks])
        f.create_dataset('padding', data=[padding])
        f.create_dataset('num_channels', data=[M])
        f.create_dataset('num_timepoints', data=[N])
        f.create_dataset('sampling_frequency', data=[recording.get_sampling_frequency()])
        f.create_dataset('channel_ids', data=channel_ids)
        f.create_dataset('channel_geom', data=geom_from_recording(recording))

        for j in range(num_chunks):
            padded_chunk = np.zeros(
                (M, chunk_size_with_padding), dtype=dtype)  # fix dtype here
            t1 = int(j*chunk_size)  # first timepoint of the chunk
            # last timepoint of chunk (+1)
            t2 = min(N, (t1+chunk_size))
            # first timepoint including the padding
            s1 = max(0, t1-padding)
            # last timepoint (+1) including the padding
            s2 = min(N, t2+padding)

            # determine aa so that t1-s1+aa = padding
            # so, aa = padding-(t1-s1)
            aa = padding-(t1-s1)
            # Read the padded chunk
            padded_chunk[:, aa:aa+s2 -
                         s1] = recording.get_traces(start_frame=s1, end_frame=s2)

            for m in range(M):
                f.create_dataset('part-{}-{}'.format(m, j),
                                 data=padded_chunk[m, :].ravel())


dummy = 0
from .averagewaveforms import *
from .clusters import *
from .correlograms import *
from .snippets import *
from .sortingcomparison import *
from .spikeamplitudes import *
from .timeseries import *
from .unitstable import *


from google.oauth2 import id_token
import cachecontrol
import google.auth.transport.requests
import requests

session = requests.session()
cached_session = cachecontrol.CacheControl(session)
request = google.auth.transport.requests.Request(session=cached_session)

def _verify_oauth2_token(token: bytes):
    id_info = id_token.verify_oauth2_token(token, request)
    return id_info


from time import time
from typing import Dict, List, Literal, Union
import kachery_client as kc
import numpy as np
from numpy.core.numeric import Infinity
from ._estimate_sampling_frequency import _estimate_sampling_frequency

class TimeseriesModelNpyV1:
    def __init__(self, *, channel_names: List[str], timestamps_uri: str, values_uri: str, type: Union[Literal['continuous'], Literal['discrete']], sampling_frequency: Union[float, None]=None, channel_properties: Union[Dict[str, dict], None]=None):
        self._channel_names = channel_names
        self._channel_properties = channel_properties
        timestamps = kc.load_npy(timestamps_uri)
        values = kc.load_npy(values_uri)
        assert timestamps is not None
        assert timestamps.ndim == 1
        assert values is not None
        assert values.ndim == 2
        assert values.shape[1] == len(channel_names)
        assert values.shape[0] == timestamps.shape[0]
        self._timestamps = timestamps
        self._values = values
        self._type: Union[Literal['continuous'], Literal['discrete']] = type
        sf: Union[float, None] = sampling_frequency
        if sf is None:
            if type == 'continuous':
                sf = _estimate_sampling_frequency(timestamps)
            else:
                sf = 0
        self._sampling_frequency: float = sf
    @property
    def channel_names(self):
        return self._channel_names
    @property
    def channel_properties(self):
        return self._channel_properties
    @property
    def num_samples(self):
        return len(self._timestamps)
    @property
    def start_time(self):
        return self._timestamps[0]
    @property
    def end_time(self):
        return self._timestamps[-1]
    @property
    def type(self) -> Union[Literal['continuous'], Literal['discrete']]:
        return self._type
    def get_samples(self, start: float, end: float, channel_inds: Union[List[int], range]):
        x = np.where((start <= self._timestamps) & (self._timestamps < end))[0]
        return (self._timestamps[x], self._values[x][:, channel_inds])
    @property
    def sampling_frequency(self):
        return self._sampling_frequency



from hdmf.utils import docval, call_docval_func, get_docval
from pynwb import register_class
from pynwb.core import NWBDataInterface


@register_class('AssociatedFiles', 'ndx-franklab-novela')
class AssociatedFiles(NWBDataInterface):
    """ Representation of associated files in NWB """

    __nwbfields__ = ('description', 'content')

    @docval(*get_docval(NWBDataInterface.__init__) + (
            {'name': 'description', 'type': 'str', 'doc': 'description of associated file'},
            {'name': 'content', 'type': 'str', 'doc': 'content of associated file'},
            ))
    def __init__(self, **kwargs):
        super().__init__(**{kwargs_item: kwargs[kwargs_item]
                            for kwargs_item in kwargs.copy()
                            if kwargs_item not in ['description', 'content']
                            })
        call_docval_func(super(AssociatedFiles, self).__init__, kwargs)
        self.description = kwargs['description']
        self.content = kwargs['content']




from copy import deepcopy
from os.path import basename
from typing import Union

import kachery_client as kc
from .h5extractors.h5recordingextractorv1 import H5RecordingExtractorV1
import numpy as np
import spikeextractors as se

from ._in_memory import (_random_string, get_in_memory_object,
                         register_in_memory_object)
from ._spikeinterface_recording_dict_to_labbox_dict import spikeinterface_recording_dict_to_labbox_dict
from .bandpass_filter import bandpass_filter
from .binextractors import Bin1RecordingExtractor, Bin2RecordingExtractor
from .mdaextractors import MdaRecordingExtractor
from .snippetsextractors import Snippets1RecordingExtractor

def _load_geom_from_csv(path: str) -> list:
    return _listify_ndarray(np.genfromtxt(path, delimiter=',').T)

def _listify_ndarray(x: np.ndarray) -> list:
    if x.ndim == 1:
        if np.issubdtype(x.dtype, np.integer):
            return [int(val) for val in x]
        else:
            return [float(val) for val in x]
    elif x.ndim == 2:
        ret = []
        for j in range(x.shape[1]):
            ret.append(_listify_ndarray(x[:, j]))
        return ret
    elif x.ndim == 3:
        ret = []
        for j in range(x.shape[2]):
            ret.append(_listify_ndarray(x[:, :, j]))
        return ret
    elif x.ndim == 4:
        ret = []
        for j in range(x.shape[3]):
            ret.append(_listify_ndarray(x[:, :, :, j]))
        return ret
    else:
        raise Exception('Cannot listify ndarray with {} dims.'.format(x.ndim))

def _try_mda_create_object(arg: Union[str, dict]) -> Union[None, dict]:
    if isinstance(arg, str):
        path = arg
        if path.startswith('sha1dir') or path.startswith('/'):
            raise Exception('sha1dir no longer supported for labbox-ephys recording extractor')
            # dd = kc.read_dir(path)
            # if dd is not None:
            #     if 'raw.mda' in dd['files'] and 'params.json' in dd['files'] and 'geom.csv' in dd['files']:
            #         raw_path = path + '/raw.mda'
            #         params_path = path + '/params.json'
            #         geom_path = path + '/geom.csv'
            #         geom_path_resolved = kc.load_file(geom_path)
            #         assert geom_path_resolved is not None, f'Unable to load geom.csv from: {geom_path}'
            #         params = kc.load_json(params_path)
            #         assert params is not None, f'Unable to load params.json from: {params_path}'
            #         geom = _load_geom_from_csv(geom_path_resolved)
            #         return dict(
            #             recording_format='mda',
            #             data=dict(
            #                 raw=raw_path,
            #                 geom=geom,
            #                 params=params
            #             )
            #         )
    
    if isinstance(arg, dict):
        if ('raw' in arg) and ('geom' in arg) and ('params' in arg) and (type(arg['geom']) == list) and (type(arg['params']) == dict):
            return dict(
                recording_format='mda',
                data=dict(
                    raw=arg['raw'],
                    geom=arg['geom'],
                    params=arg['params']
                )
            )
    
    return None

def _try_nrs_create_object(arg: Union[str, dict]) -> Union[None, dict]:
    if isinstance(arg, str):
        path = arg
        if path.startswith('sha1dir') or path.startswith('/'):
            raise Exception('sha1dir no longer supported for labbox-ephys recording extractor')
            # dd = kc.read_dir(path)
            # if dd is not None:
            #     probe_file = None
            #     xml_file = None
            #     nrs_file = None
            #     dat_file = None
            #     for f in dd['files'].keys():
            #         if f.endswith('.json'):
            #             obj = kc.load_json(path + '/' + f)
            #             if obj.get('format_version', None) in ['flatiron-probe-0.1', 'flatiron-probe-0.2']:
            #                 probe_file = path + '/' + f
            #         elif f.endswith('.xml'):
            #             xml_file = path + '/' + f
            #         elif f.endswith('.nrs'):
            #             nrs_file = path + '/' + f
            #         elif f.endswith('.dat'):
            #             dat_file = path + '/' + f
            #     if probe_file is not None and xml_file is not None and nrs_file is not None and dat_file is not None:
            #         data = dict(
            #             probe_file=probe_file,
            #             xml_file=xml_file,
            #             nrs_file=nrs_file,
            #             dat_file=dat_file
            #         )
            #         return dict(
            #             recording_format='nrs',
            #             data=data
            #         )
    
    if isinstance(arg, dict):
        if ('probe_file' in arg) and ('xml_file' in arg) and ('nrs_file' in arg) and ('dat_file' in arg):
            return dict(
                recording_format='nrs',
                data=dict(
                    probe_file=arg['probe_file'],
                    xml_file=arg['xml_file'],
                    nrs_file=arg['nrs_file'],
                    dat_file=arg['dat_file']
                )
            )
    
    return None


def _create_object_for_arg(arg: Union[str, dict]) -> Union[dict, None]:
    # if arg is a string ending with .json then replace arg by the object
    if (isinstance(arg, str)) and (arg.endswith('.json')):
        path = arg
        x = kc.load_json(path)
        if x is None:
            raise Exception(f'Unable to load object: {path}')
        return _create_object_for_arg(x)
    
    # check to see if it already has the recording_format field. If so, just return arg
    if (isinstance(arg, dict)) and ('recording_format' in arg):
        return arg

    # if has form dict(path='...') then replace by the string
    if (isinstance(arg, dict)) and ('path' in arg) and (type(arg['path']) == str):
        return _create_object_for_arg(arg['path'])

    # if has type LabboxEphysRecordingExtractor, then just get the object from arg.object()
    if isinstance(arg, LabboxEphysRecordingExtractor):
        return arg.object()

    # See if it has format 'nwb'
    if isinstance(arg, str) and arg.endswith('.nwb'):
        return dict(
            recording_format='nwb',
            data=dict(
                path=arg
            )
        )
    
    # See if it has format 'mda'
    obj = _try_mda_create_object(arg)
    if obj is not None:
        return obj
    
    # See if it has format 'nrs'
    obj = _try_nrs_create_object(arg)
    if obj is not None:
        return obj
    
    # See if it is of type filtered
    if (isinstance(arg, dict)) and ('recording' in arg) and ('filters' in arg):
        return dict(
            recording_format='filtered',
            data=dict(
                filters=arg['filters'],
                recording=_create_object_for_arg(arg['recording'])
            )
        )
    
    # See if it is type subrecording
    if (isinstance(arg, dict)) and ('recording' in arg) and ('group' in arg):
        return dict(
            recording_format='subrecording',
            data=dict(
                group=arg['group'],
                recording=_create_object_for_arg(arg['recording'])
            )
        )
    if (isinstance(arg, dict)) and ('recording' in arg) and ('groups' in arg):
        return dict(
            recording_format='subrecording',
            data=dict(
                groups=arg['groups'],
                recording=_create_object_for_arg(arg['recording'])
            )
        )
    if (isinstance(arg, dict)) and ('recording' in arg) and ('channel_ids' in arg):
        return dict(
            recording_format='subrecording',
            data=dict(
                channel_ids=arg['channel_ids'],
                recording=_create_object_for_arg(arg['recording'])
            )
        )
    
    return None    
    
# TODO: #1 reorganize this class to create a recording object that has all sha1:// paths nested
# This should be returned by the .object() method
class LabboxEphysRecordingExtractor(se.RecordingExtractor):
    def __init__(self, arg: Union[str, dict], download: bool=False):
        super().__init__()
        self.has_unscaled = False # Needed by spikeinterface?
        obj = _create_object_for_arg(arg)
        assert obj is not None
        self._object: dict = obj
        
        recording_format = self._object['recording_format']
        data: dict = self._object['data']
        if recording_format == 'mda':
            self._recording: se.RecordingExtractor = MdaRecordingExtractor(timeseries_path=data['raw'], samplerate=data['params']['samplerate'], geom=np.array(data['geom']), download=download)
        elif recording_format == 'nrs':
            self._recording: se.RecordingExtractor = NrsRecordingExtractor(**data)
        elif recording_format == 'nwb':
            from .nwbextractors import NwbRecordingExtractor
            path0 = kc.load_file(data['path'])
            self._recording: se.RecordingExtractor = NwbRecordingExtractor(path0, electrical_series_name=data.get('electrical_series_name', None))
        elif recording_format == 'bin1':
            self._recording: se.RecordingExtractor = Bin1RecordingExtractor(**data, p2p=True)
        elif recording_format == 'bin2':
            self._recording: se.RecordingExtractor = Bin2RecordingExtractor(**data)
        elif recording_format == 'snippets1':
            self._recording: se.RecordingExtractor = Snippets1RecordingExtractor(snippets_h5_uri=data['snippets_h5_uri'], p2p=True)
        elif recording_format == 'h5_v1':
            h5_uri = data['h5_uri']
            h5_path = kc.load_file(h5_uri)
            if h5_path is None:
                raise Exception(f'Unable to load h5 recording file: {h5_uri}')
            self._recording: se.RecordingExtractor = H5RecordingExtractorV1(h5_path=h5_path)
        elif recording_format == 'subrecording':
            R = LabboxEphysRecordingExtractor(data['recording'], download=download)
            if 'channel_ids' in data:
                channel_ids = np.array(data['channel_ids']) if data['channel_ids'] is not None else None
            elif 'group' in data:
                channel_ids = np.array(R.get_channel_ids())
                groups = R.get_channel_groups(channel_ids=R.get_channel_ids())
                group = int(data['group'])
                inds = np.where(np.array(groups) == group)[0]
                channel_ids = channel_ids[inds]
            elif 'groups' in data:
                raise Exception('This case not yet handled.')
            else:
                channel_ids = None
            if 'start_frame' in data:
                start_frame = data['start_frame']
                end_frame = data['end_frame']
            else:
                start_frame = None
                end_frame = None
            self._recording: se.RecordingExtractor = se.SubRecordingExtractor(
                parent_recording=R,
                channel_ids=channel_ids,
                start_frame=start_frame,
                end_frame=end_frame
            )
        elif recording_format == 'filtered':
            R = LabboxEphysRecordingExtractor(data['recording'], download=download)
            self._recording: se.RecordingExtractor = _apply_filters(recording=R, filters=data['filters'])
        elif recording_format == 'in_memory':
            R = get_in_memory_object(data)
            if R is None:
                raise Exception('Unable to find in-memory object for recording')
            self._recording = R
        else:
            raise Exception(f'Unexpected recording format: {recording_format}')

        self.copy_channel_properties(recording=self._recording)
    
    @staticmethod
    def can_load(arg):
        try:
            obj = _create_object_for_arg(arg)
        except:
            obj = None
        return (obj is not None)

    def object(self) -> dict:
        return deepcopy(self._object)
    
    def is_local(self):
        return _all_files_are_local_in_item(self._object)
    
    def download(self):
        return _download_files_in_item(self._object)

    def get_channel_ids(self) -> Union[list, np.ndarray]:
        return self._recording.get_channel_ids()

    def get_num_frames(self) -> int:
        return int(self._recording.get_num_frames())

    def get_sampling_frequency(self) -> float:
        return float(self._recording.get_sampling_frequency())

    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True) -> np.ndarray:
        return self._recording.get_traces(channel_ids=channel_ids, start_frame=start_frame, end_frame=end_frame)
    
    @staticmethod
    def from_memory(recording: se.RecordingExtractor, serialize=False, serialize_dtype=None):
        if serialize:
            if serialize_dtype is None:
                raise Exception('You must specify the serialize_dtype when serializing recording extractor in from_memory()')
            with kc.TemporaryDirectory() as tmpdir:
                fname = tmpdir + '/' + _random_string(10) + '_recording.dat'
                # se.BinDatRecordingExtractor.write_recording(recording=recording, save_path=fname, time_axis=0, dtype=serialize_dtype)
                # with ka.config(use_hard_links=True):
                recording.get_traces().T.astype(serialize_dtype).tofile(fname)
                uri = kc.store_file(fname, basename='raw.dat')
                num_channels = recording.get_num_channels()
                channel_ids = [int(a) for a in recording.get_channel_ids()]
                xcoords = [recording.get_channel_property(a, 'location')[0] for a in channel_ids]
                ycoords = [recording.get_channel_property(a, 'location')[1] for a in channel_ids]
                recording = LabboxEphysRecordingExtractor({
                    'recording_format': 'bin2',
                    'data': {
                        'raw': uri,
                        'dtype': serialize_dtype,
                        'raw_num_channels': num_channels,
                        'num_frames': int(recording.get_num_frames()),
                        'samplerate': float(recording.get_sampling_frequency()),
                        'channel_ids': channel_ids,
                        'channel_map': dict(zip([str(c) for c in channel_ids], [int(i) for i in range(num_channels)])),
                        'channel_positions': dict(zip([str(c) for c in channel_ids], [[float(xcoords[i]), float(ycoords[i])] for i in range(num_channels)]))
                    }
                })
                return recording
        obj = {
            'recording_format': 'in_memory',
            'data': register_in_memory_object(recording)
        }
        return LabboxEphysRecordingExtractor(obj)
    
    @staticmethod
    def from_spikeinterface(recording: se.RecordingExtractor):
        if not recording.is_dumpable:
            raise Exception('Cannot load recording extractor: recording is not dumpable')
        x = spikeinterface_recording_dict_to_labbox_dict(recording.dump_to_dict())
        return LabboxEphysRecordingExtractor(x)
    
    @staticmethod
    def store_recording_h5(recording: se.RecordingExtractor, format='h5_v1', dtype=float):
        if isinstance(recording, LabboxEphysRecordingExtractor):
            if recording.object()['recording_format'] == format:
                # already in this format
                print(f'Already has format: {format}')
                return recording
        if format == 'h5_v1':    
            with kc.TemporaryDirectory() as tmpdir:
                fname = tmpdir + '/recording.h5'
                print('Creating efficient recording: writing as h5...')
                H5RecordingExtractorV1.write_recording(recording=recording, h5_path=fname, dtype=dtype)
                print('Creating efficient recording: storing in kachery...')
                h5_uri = kc.store_file(fname)
                object = {
                    'recording_format': 'h5_v1',
                    'data': {
                        'h5_uri': h5_uri
                    }
                }
                print('Done creating efficient recording.')
                return LabboxEphysRecordingExtractor(object)
        else:
            raise Exception(f'Unsupported format for create_efficient_recording: {format}')
    
    @staticmethod
    def store_recording_link_h5(recording: se.RecordingExtractor, save_path:str, dtype=float):
        #print('Creating h5 recording')
        H5RecordingExtractorV1.write_recording(recording=recording, h5_path=save_path, dtype=dtype)
        print('Creating efficient recording: storing link in kachery...')
        h5_uri = kc.link_file(save_path)
        object = {
            'recording_format': 'h5_v1',
            'data': {
                'h5_uri': h5_uri
            }
        }
        print('Done creating efficient recording.')
        return LabboxEphysRecordingExtractor(object)

    # @staticmethod
    # def get_recording_object(recording):
    #     with kc.TemporaryDirectory() as tmpdir:
    #         MdaRecordingExtractor.write_recording(recording=recording, save_path=tmpdir)
    #         raw = ka.store_file(tmpdir + '/raw.mda')
    #         params = ka.load_object(tmpdir + '/params.json')
    #         geom = np.genfromtxt(tmpdir + '/geom.csv', delimiter=',').tolist()
    #         return dict(
    #             recording_format='mda',
    #             data=dict(
    #                 raw=raw,
    #                 params=params,
    #                 geom=geom
    #             )
    #         )

def _apply_filters(*, recording: se.RecordingExtractor, filters: list) -> se.RecordingExtractor:
    ret = recording
    for filter0 in filters:
        ret = _apply_filter(recording=ret, filter=filter0)
    return ret

def _apply_filter(*, recording: se.RecordingExtractor, filter: dict) -> se.RecordingExtractor:
    if filter['type'] == 'bandpass_filter':
        args = dict()
        if 'freq_min' in filter:
            args['freq_min'] = filter['freq_min']
        if 'freq_max' in filter:
            args['freq_max'] = filter['freq_max']
        if 'freq_wid' in filter:
            args['freq_wid'] = filter['freq_wid']
        return bandpass_filter(recording, **args)
    return None

class NrsRecordingExtractor(se.RecordingExtractor):
    extractor_name = 'NrsRecordingExtractor'
    is_writable = False
    def __init__(self, probe_file, xml_file, nrs_file, dat_file):
        se.RecordingExtractor.__init__(self)
        # info = check_load_nrs(dirpath)
        # assert info is not None
        probe_obj = kc.load_json(probe_file)
        xml_file = kc.load_file(xml_file)
        # nrs_file = kc.load_file(nrs_file)
        dat_file = kc.load_file(dat_file)

        from xml.etree import ElementTree as ET
        xml = ET.parse(xml_file)
        root_element = xml.getroot()
        try:
            txt = root_element.find('acquisitionSystem/samplingRate').text
            assert txt is not None
            self._samplerate = float(txt)
        except:
            raise Exception('Unable to load acquisitionSystem/samplingRate')
        try:
            txt = root_element.find('acquisitionSystem/nChannels').text
            assert txt is not None
            self._nChannels = int(txt)
        except:
            raise Exception('Unable to load acquisitionSystem/nChannels')
        try:
            txt = root_element.find('acquisitionSystem/nBits').text
            assert txt is not None
            self._nBits = int(txt)
        except:
            raise Exception('Unable to load acquisitionSystem/nBits')

        if self._nBits == 16:
            dtype = np.int16
        elif self._nBits == 32:
            dtype = np.int32
        else:
            raise Exception(f'Unexpected nBits: {self._nBits}')

        self._rec = se.BinDatRecordingExtractor(dat_file, sampling_frequency=self._samplerate, numchan=self._nChannels, dtype=dtype)

        self._channel_ids = probe_obj['channel']
        for ii in range(len(probe_obj['channel'])):
            channel = probe_obj['channel'][ii]
            x = probe_obj['x'][ii]
            y = probe_obj['y'][ii]
            z = probe_obj['z'][ii]
            group = probe_obj.get('group', probe_obj.get('shank'))[ii]
            self.set_channel_property(channel, 'location', [x, y, z])
            self.set_channel_property(channel, 'group', group)

    def get_channel_ids(self):
        return self._channel_ids

    def get_num_frames(self):
        return self._rec.get_num_frames()

    def get_sampling_frequency(self):
        return self._rec.get_sampling_frequency()

    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        if channel_ids is None:
            channel_ids = self._channel_ids
        return self._rec.get_traces(channel_ids=channel_ids, start_frame=start_frame, end_frame=end_frame)

def _all_files_are_local_in_item(x):
    if type(x) == str:
        if x.startswith('sha1://') or x.startswith('sha1dir://'):
            if kc.load_file(x):
                return False
        return True
    elif type(x) == dict:
        for _, val in x.items():
            if not _all_files_are_local_in_item(val):
                return False
        return True
    elif type(x) == list:
        for y in x:
            if not _all_files_are_local_in_item(y):
                return False
        return True
    elif type(x) == tuple:
        for y in x:
            if not _all_files_are_local_in_item(y):
                return False
        return True
    else:
        return True

def _download_files_in_item(x):
    if type(x) == str:
        if x.startswith('sha1://') or x.startswith('sha1dir://'):
            if kc.load_file(x) is None:
                a = kc.load_file(x)
                assert a is not None, f'Unable to download file: {x}'
        return
    elif type(x) == dict:
        for _, val in x.items():
            _download_files_in_item(val)
        return
    elif type(x) == list:
        for y in x:
            _download_files_in_item(y)
        return
    elif type(x) == tuple:
        for y in x:
            _download_files_in_item(y)
        return
    else:
        return



import kachery_client as kc
from ..workspace import load_workspace

@kc.taskfunction('sortingview_workspace_action.1', type='action')
def task_sortingview_workspace_action(workspace_uri: str, action: dict):
    W = load_workspace(workspace_uri)
    W._append_action(action)


from typing import Dict, List, Union

import os
import hither2 as hi
import kachery_client as kc
import numpy as np
from sortingview.config import job_cache, job_handler
from sortingview.extractors.labboxephysrecordingextractor import LabboxEphysRecordingExtractor
from sortingview.extractors.labboxephyssortingextractor import LabboxEphysSortingExtractor
from sortingview.serialize_wrapper import serialize_wrapper
from sortingview.helpers import get_unit_waveforms_from_snippets_h5
from sortingview.serialize_wrapper import _deserialize


def _compute_peak_channel_index_from_average_waveform(average_waveform):
    channel_maxes = np.max(average_waveform, axis=1)
    channel_mins = np.min(average_waveform, axis=1)
    channel_amplitudes = channel_maxes - channel_mins
    peak_channel_index = np.argmax(channel_amplitudes)
    return peak_channel_index

@hi.function(
    'fetch_spike_amplitudes', '0.1.6',
    image=hi.RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
@serialize_wrapper
def fetch_spike_amplitudes(snippets_h5: str, unit_id: Union[int, List[int]]):
    import h5py
    h5_path = kc.load_file(snippets_h5)
    assert h5_path is not None
    # with h5py.File(h5_path, 'r') as f:
    #     unit_spike_train = np.array(f.get(f'unit_spike_trains/{unit_id}'))
    #     unit_waveforms = np.array(f.get(f'unit_waveforms/{unit_id}/waveforms'))    
        
    unit_waveforms, unit_waveforms_channel_ids, channel_locations0, sampling_frequency, unit_spike_train = get_unit_waveforms_from_snippets_h5(h5_path, unit_id)
    average_waveform = np.mean(unit_waveforms, axis=0)
    peak_channel_index = _compute_peak_channel_index_from_average_waveform(average_waveform)
    maxs = [np.max(unit_waveforms[i][peak_channel_index, :]) for i in range(unit_waveforms.shape[0])]
    mins = [np.min(unit_waveforms[i][peak_channel_index, :]) for i in range(unit_waveforms.shape[0])]
    peak_amplitudes = np.array([maxs[i] - mins[i] for i in range(len(mins))])

    timepoints = unit_spike_train.astype(np.float32)
    amplitudes = peak_amplitudes.astype(np.float32)

    sort_inds = np.argsort(timepoints)
    timepoints = timepoints[sort_inds]
    amplitudes = amplitudes[sort_inds]
    
    return dict(
        timepoints=timepoints,
        amplitudes=amplitudes
    )

@kc.taskfunction('fetch_spike_amplitudes.1', type='pure-calculation')
def task_fetch_spike_amplitudes(recording_object, sorting_object, unit_id: int, snippet_len=(50, 80)):
    # Note that although unit_id can be a list, that is meant to handle the case of a merge group (a few units have been merged together)
    # For proper caching, the intent is to call this task once for every unit (or merge group)
    from sortingview.helpers import prepare_snippets_h5
    with hi.Config(job_handler=job_handler.misc, job_cache=job_cache):
        with hi.Config(job_handler=job_handler.extract_snippets):
            snippets_h5 = prepare_snippets_h5.run(recording_object=recording_object, sorting_object=sorting_object, snippet_len=snippet_len)
        return fetch_spike_amplitudes.run(
            snippets_h5=snippets_h5,
            unit_id=unit_id
        )

def runtask_fetch_spike_amplitudes(*, recording: LabboxEphysRecordingExtractor, sorting: LabboxEphysSortingExtractor, unit_id: int, snippet_len=(50, 80)):
    job = task_fetch_spike_amplitudes(
        recording_object=recording.object(),
        sorting_object=sorting.object(),
        snippet_len=snippet_len,
        unit_id=unit_id
    )
    return _deserialize(job.wait().return_value)


from .SpikeSortingView import SpikeSortingView

class SSVSegment:
    def __init__(self, *,
        parent: SpikeSortingView,
        start_frame: int,
        end_frame: int
    ) -> None:
        self._parent = parent
        self._start_frame = start_frame
        self._end_frame = end_frame
    @property
    def recording(self):
        return self._parent.recording
    @property
    def sorting(self):
        return self._parent.sorting
    @property
    def duration_minutes(self):
        return (self._end_frame - self._start_frame) / self.sampling_frequency / 60
    @property
    def start_frame(self):
        return self._start_frame
    @property
    def end_frame(self):
        return self._end_frame
    @property
    def start_minute(self):
        return self.start_frame / self.sampling_frequency / 60
    @property
    def end_minute(self):
        return self.end_frame / self.sampling_frequency / 60
    @property
    def unit_ids(self):
        return self._parent.unit_ids
    @property
    def channel_ids(self):
        return self._parent.channel_ids
    @property
    def sampling_frequency(self):
        return self._parent.sampling_frequency


import os
import stat
import numpy as np
from copy import deepcopy
from typing import Any, Dict, List, Union, cast

import kachery_client as kc
import spikeextractors as se
from spikeextractors.extractors.numpyextractors.numpyextractors import NumpySortingExtractor
from spikeextractors.sortingextractor import SortingExtractor

from sortingview.extractors.labboxephysrecordingextractor import LabboxEphysRecordingExtractor

from ._in_memory import (_random_string, get_in_memory_object,
                         register_in_memory_object)
from .h5extractors.h5sortingextractorv1 import H5SortingExtractorV1
from .mdaextractors import MdaSortingExtractor
from .snippetsextractors import Snippets1SortingExtractor
from .curatedsortingextractor import CuratedSortingExtractor


def _try_mda_create_object(arg: Union[str, dict], samplerate=None) -> Union[None, dict]:
    if isinstance(arg, str):
        path = arg
        if not kc.load_file(path):
            return None
        return dict(
            sorting_format='mda',
            data=dict(
                firings=path,
                samplerate=samplerate
            )
        )
    
    if isinstance(arg, dict):
        if 'firings' in arg:
            return dict(
                sorting_format='mda',
                data=dict(
                    firings=arg['firings'],
                    samplerate=arg.get('samplerate', 30000)
                )
            )
    
    return None

def _create_object_for_arg(arg: Union[str, dict], samplerate=None) -> Union[dict, None]:
    # check to see if it already has the sorting_format field. If so, just return arg
    if (isinstance(arg, dict)) and ('sorting_format' in arg):
        return arg

    # if has form dict(path='...') then replace by the string
    if (isinstance(arg, dict)) and ('path' in arg) and (type(arg['path']) == str):
        arg = arg['path']

    # if has type LabboxEphysSortingExtractor, then just get the object from arg.object()
    if isinstance(arg, LabboxEphysSortingExtractor):
        return arg.object()

    # if arg is a string ending with .json then replace arg by the object
    if (isinstance(arg, str)) and (arg.endswith('.json')):
        path = arg
        obj = kc.load_json(path)
        if obj is None:
            raise Exception(f'Unable to load json object: {path}')
        return obj
    
    # See if it has format 'mda'
    obj = _try_mda_create_object(arg, samplerate=samplerate)
    if obj is not None:
        return obj
    
    return None    

class LabboxEphysSortingExtractor(se.SortingExtractor):
    def __init__(self, arg, samplerate=None):
        super().__init__()
        if (isinstance(arg, dict)) and ('sorting_format' in arg):
            obj = dict(arg)
        else:
            obj = _create_object_for_arg(arg, samplerate=samplerate)
            assert obj is not None, f'Unable to create sorting from arg: {arg}'
        self._object: dict = obj

        if 'firings' in self._object:
            sorting_format = 'mda'
            data: Dict[str, Any]={'firings': self._object['firings'], 'samplerate': self._object.get('samplerate', 30000)}
        else:
            sorting_format = self._object['sorting_format']
            data: Dict[str, Any] = self._object['data']
        if sorting_format == 'mda':
            firings_path = kc.load_file(data['firings'])
            assert firings_path is not None, f'Unable to load firings file: {data["firings"]}'
            self._sorting: se.SortingExtractor = MdaSortingExtractor(firings_file=firings_path, samplerate=data['samplerate'])
        elif sorting_format == 'h5_v1':
            h5_path = kc.load_file(data['h5_path'])
            self._sorting = H5SortingExtractorV1(h5_path=h5_path)
        elif sorting_format == 'npy1':
            times_npy = kc.load_npy(data['times_npy_uri'])
            labels_npy = kc.load_npy(data['labels_npy_uri'])
            samplerate = data['samplerate']
            S = se.NumpySortingExtractor()
            S.set_sampling_frequency(samplerate)
            S.set_times_labels(times_npy.ravel(), labels_npy.ravel())
            self._sorting = S
        elif sorting_format == 'snippets1':
            S = Snippets1SortingExtractor(snippets_h5_uri = data['snippets_h5_uri'], p2p=True)
            self._sorting = S
        elif sorting_format == 'npy2':
            npz = kc.load_npy(data['npz_uri'])
            times_npy = npz['spike_indexes']
            labels_npy = npz['spike_labels']
            samplerate = float(npz['sampling_frequency'])
            S = se.NumpySortingExtractor()
            S.set_sampling_frequency(samplerate)
            S.set_times_labels(times_npy.ravel(), labels_npy.ravel())
            self._sorting = S
        elif sorting_format == 'nwb':
            from .nwbextractors import NwbSortingExtractor
            path0 = kc.load_file(data['path'])
            self._sorting: se.SortingExtractor = NwbSortingExtractor(path0)
        elif sorting_format == 'in_memory':
            S = get_in_memory_object(data)
            if S is None:
                raise Exception('Unable to find in-memory object for sorting')
            self._sorting = S
        elif sorting_format == 'curated':
            parent_sorting = LabboxEphysSortingExtractor(data['sorting'])
            merge_groups = data.get('merge_groups', [])
            S = CuratedSortingExtractor(parent_sorting=parent_sorting, merge_groups=merge_groups)
            self._sorting = S
        elif sorting_format == 'subsorting':
            parent_sorting = LabboxEphysSortingExtractor(data['sorting'])
            start_frame = data.get('start_frame', None)
            end_frame = data.get('end_frame', None)
            unit_ids = data.get('unit_ids', None)
            self._sorting = se.SubSortingExtractor(parent_sorting=parent_sorting, unit_ids=unit_ids, start_frame=start_frame, end_frame=end_frame)
        else:
            raise Exception(f'Unexpected sorting format: {sorting_format}')

        self.copy_unit_properties(sorting=self._sorting)
    
    def object(self):
        return deepcopy(self._object)

    def get_unit_ids(self) -> List[int]:
        return cast(List[int], self._sorting.get_unit_ids())

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        return self._sorting.get_unit_spike_train(unit_id=unit_id, start_frame=start_frame, end_frame=end_frame)
    
    def get_sampling_frequency(self):
        return self._sorting.get_sampling_frequency()
    
    def set_sampling_frequency(self, freq):
        self._sorting.set_sampling_frequency(freq)
    
    @staticmethod
    def from_numpy(samplerate: float, times: np.array, labels: np.array):
        with kc.TemporaryDirectory() as tmpdir:
            h5_path = tmpdir + '/sorting.h5'
            S = NumpySortingExtractor()
            S.set_sampling_frequency(samplerate)
            S.set_times_labels(times, labels)
            H5SortingExtractorV1.write_sorting(sorting=S, save_path=h5_path)
            return LabboxEphysSortingExtractor({
                'sorting_format': 'h5_v1',
                'data': {
                    'h5_path': kc.store_file(h5_path)
                }
            })

    @staticmethod
    def from_memory(sorting: se.SortingExtractor, serialize=False):
        if serialize:
            with kc.TemporaryDirectory() as tmpdir:
                fname = tmpdir + '/' + _random_string(10) + '_firings.mda'
                MdaSortingExtractor.write_sorting(sorting=sorting, save_path=fname)
                # with ka.config(use_hard_links=True):
                uri = kc.store_file(fname, basename='firings.mda')
                sorting = LabboxEphysSortingExtractor({
                    'sorting_format': 'mda',
                    'data': {
                        'firings': uri,
                        'samplerate': sorting.get_sampling_frequency()
                    }
                })
                return sorting
        obj = {
            'sorting_format': 'in_memory',
            'data': register_in_memory_object(sorting)
        }
        return LabboxEphysSortingExtractor(obj)
    
    @staticmethod
    def write_sorting(sorting, save_path=None):
        if save_path is not None:
            print('WARNING: save_path not used in LabboxEphysSortingExtractor.write_sorting')
        with kc.TemporaryDirectory() as tmpdir:
            H5SortingExtractorV1.write_sorting(sorting=sorting, save_path=tmpdir + '/' + _random_string(10) + '_sorting.h5')
    
    @staticmethod
    def store_sorting(sorting: SortingExtractor):
        with kc.TemporaryDirectory() as tmpdir:
            save_path = tmpdir + '/sorting.h5'
            H5SortingExtractorV1.write_sorting(sorting=sorting, save_path=save_path)

            # in case we are in a container, the daemon needs to be able to access this file
            _add_read_permissions(tmpdir)
            _add_exec_permissions(tmpdir)
            _add_read_permissions(save_path)

            uri =kc.store_file(save_path)
            object = {
                'sorting_format': 'h5_v1',
                'data': {
                    'h5_path': uri
                }
            }
            return LabboxEphysSortingExtractor(object)

    @staticmethod
    def store_sorting_link_h5(sorting: SortingExtractor, save_path: str):
        H5SortingExtractorV1.write_sorting(sorting=sorting, save_path=save_path)
        h5_uri = kc.link_file(save_path)
        object = {
            'sorting_format': 'h5_v1',
            'data': {
                'h5_path': h5_uri
            }
        }
        return LabboxEphysSortingExtractor(object)

def _add_read_permissions(fname: str):
    st = os.stat(fname)
    os.chmod(fname, st.st_mode | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)

def _add_exec_permissions(fname: str):
    st = os.stat(fname)
    os.chmod(fname, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)



from functools import wraps 
import base64
import numpy as np

def serialize_wrapper(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        output = f(*args, **kwargs)
        return _serialize(output)
    return wrapper

def _serialize(x):
    if isinstance(x, np.integer):
        return int(x)
    elif isinstance(x, np.floating):
        return float(x)
    elif type(x) == dict:
        ret = dict()
        for key, val in x.items():
            ret[key] = _serialize(val)
        return ret
    elif (type(x) == list) or (type(x) == tuple):
        return [_serialize(val) for val in x]
    elif isinstance(x, np.ndarray):
        # todo: worry about byte order and data type here
        return {
            '_type': 'ndarray',
            'shape': _serialize(x.shape),
            'dtype': str(x.dtype),
            'data_b64': base64.b64encode(x.ravel()).decode()
        }
    else:
        if _is_jsonable(x):
            # this will capture int, float, str, bool
            return x
    raise Exception(f'Item is not json safe: {type(x)}')

def _deserialize(x):
    if type(x) == dict:
        if x.get('_type', None) == 'ndarray':
            shape = x['shape']
            dtype = x['dtype']
            data_b64 = x['data_b64']
            data = base64.b64decode(data_b64)
            x = np.reshape(np.frombuffer(data, dtype=dtype), shape)
            return x
        else:
            ret = dict()
            for key, val in x.items():
                ret[key] = _deserialize(val)
            return ret
    elif (type(x) == list) or (type(x) == tuple):
        return [_deserialize(val) for val in x]
    else:
        return x

def _is_jsonable(x) -> bool:
    import json
    try:
        json.dumps(x)
        return True
    except:
        return False


import spikeextractors as se
import numpy as np
import time

def _extract_snippet_from_traces(
    traces,
    start_frame,
    end_frame,
    channel_indices
):
    if (0 <= start_frame) and (end_frame <= traces.shape[1]):
        x = traces[:, start_frame:end_frame]
    else:
        # handle edge cases
        x = np.zeros((traces.shape[0], end_frame - start_frame), dtype=traces.dtype)
        i1 = int(max(0, start_frame))
        i2 = int(min(traces.shape[1], end_frame))
        x[:, (i1 - start_frame):(i2 - start_frame)] = traces[:, i1:i2]
    if channel_indices is not None:
        x = x[channel_indices, :]
    return x

def _get_unit_waveforms_for_chunk(
    recording,
    sorting,
    frame_offset,
    unit_ids,
    snippet_len,
    channel_ids_by_unit
):
    # chunks are chosen small enough so that all traces can be loaded into memory
    print('Retrieving traces for chunk')
    traces = recording.get_traces()

    print('Collecting waveforms for chunk')
    unit_waveforms = []
    for unit_id in unit_ids:
        times0 = sorting.get_unit_spike_train(unit_id=unit_id)
        if channel_ids_by_unit is not None:
            channel_ids = channel_ids_by_unit[unit_id]
            all_channel_ids = recording.get_channel_ids()
            channel_indices = [
                np.array(all_channel_ids).tolist().index(ch_id)
                for ch_id in channel_ids
            ]
            len_channel_indices = len(channel_indices)
        else:
            channel_indices = None
            len_channel_indices = traces.shape[0]
        # num_channels x len_of_one_snippet
        if len(times0) > 0:
            snippets = [
                _extract_snippet_from_traces(
                    traces,
                    start_frame=frame_offset + int(t) - snippet_len[0],
                    end_frame=frame_offset + int(t) + snippet_len[1],
                    channel_indices=channel_indices
                )
                for t in times0
            ]
            unit_waveforms.append(
                # len(times0) x num_channels_in_nbhd[unit_id] x len_of_one_snippet
                np.stack(snippets)
            )
        else:
            unit_waveforms.append(
                np.zeros((0, len_channel_indices, snippet_len[0] + snippet_len[1]), dtype=traces.dtype)
            )
    return unit_waveforms

def _divide_recording_into_time_chunks(num_frames, chunk_size, padding_size):
    chunks = []
    ii = 0
    while ii < num_frames:
        ii2 = int(min(ii + chunk_size, num_frames))
        chunks.append(dict(
            istart=ii,
            iend=ii2,
            istart_with_padding=int(max(0, ii - padding_size)),
            iend_with_padding=int(min(num_frames, ii2 + padding_size))
        ))
        ii = ii2
    return chunks

def get_unit_waveforms(
    recording,
    sorting,
    unit_ids,
    channel_ids_by_unit,
    snippet_len
):
    if not isinstance(snippet_len, list) and not isinstance(snippet_len, tuple):
        b = int(snippet_len / 2)
        a = int(snippet_len) - b
        snippet_len = [a, b]

    num_channels = recording.get_num_channels()
    num_frames = recording.get_num_frames()
    num_bytes_per_chunk = 1000 * 1000 * 1000 # ? how to choose this
    num_bytes_per_frame = num_channels * 2
    chunk_size = num_bytes_per_chunk / num_bytes_per_frame
    padding_size = 100 + snippet_len[0] + snippet_len[1] # a bit excess padding
    chunks = _divide_recording_into_time_chunks(
        num_frames=num_frames,
        chunk_size=chunk_size,
        padding_size=padding_size
    )
    all_unit_waveforms = [[] for ii in range(len(unit_ids))]
    for ii, chunk in enumerate(chunks):
        # chunk: {istart, iend, istart_with_padding, iend_with_padding} # include padding
        print(f'Processing chunk {ii + 1} of {len(chunks)}; chunk-range: {chunk["istart_with_padding"]} {chunk["iend_with_padding"]}; num-frames: {num_frames}')
        recording_chunk = se.SubRecordingExtractor(
            parent_recording=recording,
            start_frame=chunk['istart_with_padding'],
            end_frame=chunk['iend_with_padding']
        )
        # note that the efficiency of this operation may need improvement (really depends on sorting extractor implementation)
        sorting_chunk = se.SubSortingExtractor(
            parent_sorting=sorting,
            start_frame=chunk['istart'],
            end_frame=chunk['iend']
        )
        print(f'Getting unit waveforms for chunk {ii + 1} of {len(chunks)}')
        # num_events_in_chunk x num_channels_in_nbhd[unit_id] x len_of_one_snippet
        unit_waveforms = _get_unit_waveforms_for_chunk(
            recording=recording_chunk,
            sorting=sorting_chunk,
            frame_offset=chunk['istart'] - chunk['istart_with_padding'], # just the padding size (except 0 for first chunk)
            unit_ids=unit_ids,
            snippet_len=snippet_len,
            channel_ids_by_unit=channel_ids_by_unit
        )
        for i_unit, x in enumerate(unit_waveforms):
            all_unit_waveforms[i_unit].append(x)
    
    # concatenate the results over the chunks
    unit_waveforms = [
        # tot_num_events_for_unit x num_channels_in_nbhd[unit_id] x len_of_one_snippet
        np.concatenate(all_unit_waveforms[i_unit], axis=0)
        for i_unit in range(len(unit_ids))
    ]
    return unit_waveforms



import kachery_client as kc
import hither2 as hi
from sortingview.config import job_cache, job_handler
from sortingview.extractors import LabboxEphysRecordingExtractor, LabboxEphysSortingExtractor

@hi.function(
    'get_best_matching_units', '0.1.2',
    image=hi.RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
def get_best_matching_units(*, sorting_object, compare_sorting_object, sorting_selector, recording_object):
    import spikecomparison as sc
    S = LabboxEphysSortingExtractor(sorting_object)
    S_compare = LabboxEphysSortingExtractor(compare_sorting_object)
    R = LabboxEphysRecordingExtractor(recording_object)
    C = sc.compare_sorter_to_ground_truth(gt_sorting=S, tested_sorting=S_compare)
    if sorting_selector == 'A':
        s = 'B'
    elif sorting_selector == 'B':
        s = 'A'
    else:
        s = ''
    ret = {}
    for id in S.get_unit_ids():
        ret[str(id)] = f'{C.best_match_12[id]}{s}'
    return ret
    


@kc.taskfunction('get_best_matching_units.2', type='pure-calculation')
def task_get_best_matching_units(*, sorting_object, compare_sorting_object, sorting_selector, recording_object, configuration={}, snippet_len=(50, 80)):
    with hi.Config(
        job_cache=job_cache,
        job_handler=job_handler.metrics
    ):
        return get_best_matching_units.run(
            sorting_object=sorting_object,
            compare_sorting_object=compare_sorting_object,
            sorting_selector=sorting_selector,
            recording_object=recording_object
        )


import os
import kachery_client as kc
from .workspace import Workspace
from ..tasks.preload_extract_snippets import task_preload_extract_snippets
from sortingview.helpers import prepare_snippets_h5

def _precalculate(workspace: Workspace):
    FIGURL_CHANNEL = os.getenv('FIGURL_CHANNEL')
    if FIGURL_CHANNEL is None:
        raise Exception('Environment variable not set: FIGURL_CHANNEL')
    for sorting_id in workspace.sorting_ids:
        sorting = workspace.get_sorting(sorting_id)
        sorting_id = sorting['sortingId']
        sorting_label = sorting['sortingLabel']
        sorting_object = sorting['sortingObject']
        recording_id = sorting['recordingId']
        recording_object = sorting['recordingObject']
        recording = workspace.get_recording(recording_id)
        recording_label = recording['recordingLabel']

        print(f'Precalculating: {recording_label}/{sorting_label}')
        kwargs = {
            'recording_object': recording_object,
            'sorting_object': sorting_object
        }
        if workspace.snippet_len != (50, 80):
            # important to only add this kwarg if not the default
            kwargs['snippet_len'] = workspace.snippet_len
        kc._run_task(
            task_preload_extract_snippets,
            kwargs,
            channel=FIGURL_CHANNEL
        )

def _precalculate_debug(workspace: Workspace):
    for sorting_id in workspace.sorting_ids:
        sorting = workspace.get_sorting(sorting_id)
        sorting_id = sorting['sortingId']
        sorting_label = sorting['sortingLabel']
        sorting_object = sorting['sortingObject']
        recording_id = sorting['recordingId']
        recording_object = sorting['recordingObject']
        recording = workspace.get_recording(recording_id)
        recording_label = recording['recordingLabel']

        print(f'Precalculating (debug): {recording_label}/{sorting_label}')
        kwargs = {
            'recording_object': recording_object,
            'sorting_object': sorting_object
        }
        if workspace.snippet_len != (50, 80):
            # important to only add this kwarg if not the default
            kwargs['snippet_len'] = workspace.snippet_len
        
        prepare_snippets_h5(
            **kwargs
        )


import numpy as np
import kachery_client as kc
import hither2 as hi
from sortingview import experitime
from sortingview.config import job_cache, job_handler
from sortingview.experitime import timeseries
from sortingview.serialize_wrapper import serialize_wrapper

@hi.function('get_experitime_timeseries_samples', '0.1.2')
@serialize_wrapper
def get_experitime_timeseries_samples(timeseries_uri: str, channel_name: str, ds_factor: int, segment_num: int, segment_duration_sec: float):
    t1 = segment_num * segment_duration_sec
    t2 = (segment_num + 1) * segment_duration_sec
    x = experitime.Timeseries(timeseries_uri)
    timestamps, values = x.get_samples(start=t1, end=t2, channels=[channel_name])
    values = values.ravel()
    if ds_factor > 1:
        N = len(timestamps)
        N2 = int(N / ds_factor)
        timestamps = timestamps[:N2 * ds_factor]
        values = values[:N2 * ds_factor]
        timestamps_reshaped = timestamps.reshape((N2, ds_factor))
        values_reshaped = values.reshape((N2, ds_factor))
        values_min = np.min(values_reshaped, axis=1)
        values_max = np.max(values_reshaped, axis=1)        
        values = np.zeros((N2 * 2,))
        values[0::2] = values_min
        values[1::2] = values_max
        timestamps = timestamps_reshaped[:, 0].ravel()
    return {
        'timestamps': timestamps.astype(np.float32),
        'values': values.astype(np.float32)
    }
    

@kc.taskfunction('experitime.get_timeseries_samples.2', type='pure-calculation')
def task_get_timeseries_samples(timeseries_uri, channel_name: str, ds_factor: int, segment_num: int, segment_duration_sec: float):
    with hi.Config(job_handler=job_handler.misc, job_cache=job_cache):
        return hi.Job(get_experitime_timeseries_samples, {'timeseries_uri': timeseries_uri, 'channel_name': channel_name, 'ds_factor': ds_factor, 'segment_num': segment_num, 'segment_duration_sec': segment_duration_sec})

@hi.function('get_experitime_timeseries_info', '0.1.2')
@serialize_wrapper
def get_experitime_timeseries_info(timeseries_uri: str):
    x = experitime.Timeseries(timeseries_uri)
    return {
        'uri': timeseries_uri,
        'object': x.object,
        'channelProperties': x.channel_properties,
        'channelNames': x.channel_names,
        'numSamples': x.num_samples,
        'startTime': x.start_time,
        'samplingFrequency': x.sampling_frequency,
        'endTime': x.end_time,
        'type': x.type
    }

@kc.taskfunction('experitime.get_timeseries_info.2', type='pure-calculation')
def task_get_timeseries_info(timeseries_uri):
    with hi.Config(job_handler=job_handler.misc, job_cache=job_cache):
        return hi.Job(get_experitime_timeseries_info, {'timeseries_uri': timeseries_uri})


import random


def _random_string(num_chars: int) -> str:
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choice(chars) for _ in range(num_chars))

unique_process_id = _random_string(20)
in_memory_objects = {}

def register_in_memory_object(obj):
    in_memory_objects[id(obj)] = obj
    return {
        'process_id': unique_process_id,
        'object_id': id(obj)
    }

def get_in_memory_object(a: dict):
    object_id = a['object_id']
    process_id = a['process_id']
    if process_id == unique_process_id:
        if object_id in in_memory_objects:
            return in_memory_objects[object_id]
    return None



import json
from typing import Any, Dict, List, Tuple, Union
import uuid
import kachery_client as kc
from figurl import Figure
import spikeextractors as se
from ..extractors import LabboxEphysRecordingExtractor, LabboxEphysSortingExtractor

def parse_workspace_uri(workspace_uri: str):
    if not workspace_uri.startswith('workspace://'):
        raise Exception(f'Invalid workspace uri: {workspace_uri}')
    if '?' not in workspace_uri:
        workspace_uri = workspace_uri + '?'
    params = {}
    ind = workspace_uri.index('?')
    feed_id = workspace_uri[:ind].split('/')[2]
    query_string = workspace_uri[ind+1:]
    return feed_id, query_string

class Workspace:
    def __init__(self, *, workspace_uri: str, label: Union[None, str]=None) -> None:
        feed_id, query_string = parse_workspace_uri(workspace_uri)
        self._query_string = query_string
        self._feed = kc.load_feed(f'feed://{feed_id}')
        main_subfeed = self._feed.load_subfeed('main')
        messages = _get_messages_from_subfeed(main_subfeed)
        self._recordings = _get_recordings_from_subfeed_messages(messages)
        self._sortings = _get_sortings_from_subfeed_messages(messages)
        self._unit_metrics_for_sortings = _get_unit_metrics_for_sortings_from_subfeed_messages(messages)
        self._user_permissions = _get_user_permissions_from_subfeed_messages(messages)
        self._snippet_len: Tuple[int, int] = _get_snippet_len_from_subfeed_messages(messages)
        self._label = label
    @property
    def uri(self):
        q = f'?{self._query_string}' if self._query_string else ''
        return f'workspace://{self._feed.feed_id}{q}'
    @property
    def feed_uri(self):
        return self._feed.uri
    @property
    def feed_id(self):
        return self._feed.feed_id
    @property
    def feed(self):
        return self._feed
    @property
    def label(self):
        if self._label is not None:
            return self._label
        p = _query_string_to_dict(self._query_string)
        return p.get('label', '')
    def get_uri(self):
        print('WARNING: workspace.get_uri() is deprecated. Use workspace.uri instead')
        return self.uri
    def get_feed_id(self):
        print('WARNING: workspace.feed_id is deprecated. Use workspace.feed_id instead')
        return self.feed_id
    def set_label(self, label: str):
        p = _query_string_to_dict(self._query_string)
        if label:
            p['label'] = label
        else:
            if 'label' in p:
                del p['label']
        self._query_string = _dict_to_query_string(p)
    def add_recording(self, *, label: str, recording: LabboxEphysRecordingExtractor):
        recording_id = 'R-' + _random_id()
        if recording_id in self._recordings:
            raise Exception(f'Duplicate recording ID: {recording_id}')
        x = {
            'recordingId': recording_id,
            'recordingLabel': label,
            'recordingPath': kc.store_json(recording.object(), basename=f'{label}.json'),
            'recordingObject': recording.object(),
            'description': f'Imported from Python: {label}'
        }
        main_subfeed = self._feed.load_subfeed('main')
        _import_le_recording(main_subfeed, x)
        self._recordings[recording_id] = x
        return recording_id
    def add_sorting(self, *, recording_id: str, label: str, sorting: LabboxEphysSortingExtractor):
        sorting_id = 'S-' + _random_id()
        if recording_id not in self._recordings:
            raise Exception(f'Recording not found: {recording_id}')
        if sorting_id in self._sortings:
            raise Exception(f'Duplicate sorting ID: {sorting_id}')
        le_recording = self._recordings[recording_id]
        x = {
            'sortingId': sorting_id,
            'sortingLabel': label,
            'sortingPath': kc.store_json(sorting.object(), basename=f'{label}.json'),
            'sortingObject': sorting.object(),

            'recordingId': recording_id,
            'recordingPath': le_recording['recordingPath'],
            'recordingObject': le_recording['recordingObject'],

            'description': f'Imported from Python: {label}'
        }
        main_subfeed = self._feed.load_subfeed('main')
        _import_le_sorting(main_subfeed, x)
        self._sortings[sorting_id] = x
        return sorting_id
    def precalculate(self):
        from ._precalculate import _precalculate
        _precalculate(self)
    def _precalculate_debug(self):
        from ._precalculate import _precalculate_debug
        _precalculate_debug(self)
    def _append_action(self, action: dict):
        main_subfeed = self._feed.load_subfeed('main')
        main_subfeed.append_message({
            'action': action
        })
    def set_unit_metrics_for_sorting(self, *, sorting_id: str, metrics: List[dict]):
        metrics_uri = kc.store_json(metrics, basename='unit_metrics.json')
        x = {
            'sortingId': sorting_id,
            'metricsUri': metrics_uri
        }
        main_subfeed = self._feed.load_subfeed('main')
        _set_unit_metrics_for_sorting(main_subfeed, x)
        self._unit_metrics_for_sortings[sorting_id] = metrics
    def delete_recording(self, recording_id: str):
        if recording_id not in self._recordings:
            raise Exception(f'Recording not found: {recording_id}')
        _delete_recording(feed=self._feed, recording_id=recording_id)
        del self._recordings[recording_id]
    def delete_sorting(self, sorting_id: str):
        if sorting_id not in self._sortings:
            raise Exception(f'Sorting not found: {sorting_id}')
        _delete_sorting(feed=self._feed, sorting_id=sorting_id)
        del self._sortings[sorting_id]
    def set_user_permissions(self, user_id: str, permissions: dict):
        if json.dumps(permissions) == json.dumps(self._user_permissions.get(user_id, {})):
            return
        main_subfeed = self._feed.load_subfeed('main')
        _set_user_permissions_for_workspace(main_subfeed, user_id, permissions)
        self._user_permissions[user_id] = permissions
    def set_snippet_len(self, snippet_len: Tuple[int, int]):
        main_subfeed = self._feed.load_subfeed('main')
        _set_snippet_len_for_workspace(main_subfeed, snippet_len)
        self._snippet_len = snippet_len
    def figurl(self):
        data = {
            'type': 'workspace',
            'workspaceUri': self.uri
        }
        return Figure(view_url='gs://figurl/sortingview-gui-1', data=data)
    def figurl_old(self):
        data = {
            'workspaceUri': self.uri
        }
        return Figure(type='sortingview.workspace.1', data=data)
    @property
    def snippet_len(self):
        return self._snippet_len
    def get_user_permissions(self, user_id: str) -> Union[None, dict]:
        return self._user_permissions.get(user_id, None)
    def get_all_users(self) -> List[str]:
        return list(self._user_permissions.keys())
    def get_recording(self, recording_id: str):
        return self._recordings[recording_id]
    def get_sorting(self, sorting_id: str):
        return self._sortings[sorting_id]
    @property
    def recording_ids(self):
        return list(self._recordings.keys())
    @property
    def sorting_ids(self):
        return list(self._sortings.keys())
    def get_sorting_ids_for_recording(self, recording_id: str):
        return [sid for sid in self.sorting_ids if self.get_sorting(sid)['recordingId'] == recording_id]
    def get_recording_extractor(self, recording_id):
        r = self.get_recording(recording_id)
        return LabboxEphysRecordingExtractor(r['recordingObject'])
    def get_sorting_extractor(self, sorting_id):
        s = self.get_sorting(sorting_id)
        return LabboxEphysSortingExtractor(s['sortingObject'])
    def get_sorting_curation(self, sorting_id: str):
        curation_subfeed = self._feed.load_subfeed(dict(name='sortingCuration', sortingId=sorting_id))
        return _get_sorting_curation(curation_subfeed, sorting_id=sorting_id)
    def get_curated_sorting_extractor(self, sorting_id):
        s = self.get_sorting(sorting_id)
        sc = self.get_sorting_curation(sorting_id)
        return LabboxEphysSortingExtractor({
            'sorting_format': 'curated',
            'data': {
                'sorting': s['sortingObject'],
                'merge_groups': sc.get('mergeGroups', [])
            }
        })

def create_workspace(*, label: Union[str, None]=None):
    feed = kc.create_feed()
    feed_id = feed.feed_id
    workspace_uri = f'workspace://{feed_id}'
    W = load_workspace(workspace_uri)
    if label:
        W.set_label(label)
    return W

def load_workspace(workspace_uri: Union[str, Any]):
    if not isinstance(workspace_uri, str):
        raise Exception('Invalid workspace URI')
    return Workspace(workspace_uri=workspace_uri)

def _random_id():
    return str(uuid.uuid4())[-12:]

def _get_messages_from_subfeed(subfeed: kc.Subfeed):
    subfeed.set_position(0)
    messages: List[Any] = []
    while True:
        msgs = subfeed.get_next_messages(wait_msec=0)
        if msgs is None: break
        if len(msgs) == 0: break
        for msg in msgs:
            messages.append(msg)
    return messages

def _get_recordings_from_subfeed_messages(messages: List[Any]):
    le_recordings = {}
    for msg in messages:
        if 'action' in msg:
            a = msg['action']
            msg_type = a.get('type', '')
            if msg_type == 'ADD_RECORDING':
                r = a.get('recording', {})
                rid = r.get('recordingId', '')
                le_recordings[rid] = r
            elif msg_type == 'DELETE_RECORDINGS':
                for rid in a.get('recordingIds', []):
                    if rid in le_recordings:
                        del le_recordings[rid]
    return le_recordings

def _get_user_permissions_from_subfeed_messages(messages: List[Any]):
    user_permissions = {}
    for msg in messages:
        if 'action' in msg:
            a = msg['action']
            if a.get('type', '') == 'SET_USER_PERMISSIONS':
                user_id = a.get('userId', None)
                permissions = a.get('permissions', None)
                if user_id and permissions:
                    user_permissions[user_id] = permissions
    return user_permissions

def _set_user_permissions_for_workspace(subfeed: kc.Subfeed, user_id: str, permissions: dict):
    subfeed.append_message({
        'action': {
            'type': 'SET_USER_PERMISSIONS',
            'userId': user_id,
            'permissions': permissions
        }
    })

def _get_snippet_len_from_subfeed_messages(messages: List[Any]):
    snippet_len = (50, 80)
    for msg in messages:
        if 'action' in msg:
            a = msg['action']
            if a.get('type', '') == 'SET_SNIPPET_LEN':
                x = a.get('snippetLen', None)
                if x:
                    snippet_len = x
    return snippet_len

def _set_snippet_len_for_workspace(subfeed: kc.Subfeed, snippet_len: Tuple[int, int]):
    subfeed.append_message({
        'action': {
            'type': 'SET_SNIPPET_LEN',
            'snippetLen': snippet_len
        }
    })

def _get_sortings_from_subfeed_messages(messages: List[Any]):
    le_sortings = {}
    for msg in messages:
        if 'action' in msg:
            a = msg['action']
            msg_type = a.get('type', '')
            if msg_type == 'ADD_SORTING':
                s = a.get('sorting', {})
                sid = s.get('sortingId', '')
                le_sortings[sid] = s
            elif msg_type == 'DELETE_SORTINGS':
                for sid in a.get('sortingIds', []):
                    if sid in le_sortings:
                        del le_sortings[sid]
            elif msg_type == 'DELETE_SORTINGS_FOR_RECORDINGS':
                for rid in a.get('recordingIds', []):
                    sids = list(le_sortings.keys())
                    for sid in sids:
                        if le_sortings[sid]['recordingId'] == rid:
                            del le_sortings[sid]
            elif msg_type == 'DELETE_RECORDINGS':
                for rid in a.get('recordingIds', []):
                    sids = list(le_sortings.keys())
                    for sid in sids:
                        if le_sortings[sid]['recordingId'] == rid:
                            del le_sortings[sid]
    return le_sortings

def _get_unit_metrics_for_sortings_from_subfeed_messages(messages: List[Any]):
    sortings = _get_sortings_from_subfeed_messages(messages)
    le_unit_metrics_for_sortings = {}
    for msg in messages:
        if 'action' in msg:
            a = msg['action']
            if a.get('type', '') == 'SET_UNIT_METRICS_FOR_SORTING':
                x = a.get('unitMetricsForSorting', {})
                sid = x.get('sortingId', '')
                uri = x.get('metricsUri')
                if sid in sortings:
                    le_unit_metrics_for_sortings[sid] = kc.load_json(uri)
    return le_unit_metrics_for_sortings

def _mg_intersection(g1: List[int], g2: List[int]):
    return [x for x in g1 if x in g2]

def _mg_union(g1: List[int], g2: List[int]):
    return sorted(list(set(g1 + g2)))

def _simplify_merge_groups(merge_groups: List[List[int]]):
    new_merge_groups: List[List[int]] = [[x for x in g] for g in merge_groups] # make a copy
    something_changed = True
    while something_changed:
        something_changed = False
        for i in range(len(new_merge_groups)):
            g1 = new_merge_groups[i]
            for j in range(i + 1, len(new_merge_groups)):
                g2 = new_merge_groups[j]
                if len(_mg_intersection(g1, g2)) > 0:
                    new_merge_groups[i] = _mg_union(g1, g2)
                    new_merge_groups[j] = []
                    something_changed = True
    return [sorted(mg) for mg in new_merge_groups if len(mg) >= 2]

def _get_sorting_curation(subfeed: kc.Subfeed, sorting_id: str):
    subfeed.set_position(0)
    labels_by_unit = {}
    merge_groups = []
    is_closed = False
    while True:
        msgs = subfeed.get_next_messages(wait_msec=0)
        if msgs is None: break
        if len(msgs) == 0: break
        for a in msgs:
            message_type = a.get('type', None)
            assert message_type is not None, "Feed contained message with no type."
            # if is_closed and message_type != 'REOPEN_CURATION':
            #    raise Exception('ERROR: Subfeed attempts curation on a closed curation object.')
            if message_type == 'ADD_UNIT_LABEL':
                unit_ids = a.get('unitId', []) # allow this to be a list or an int
                if not isinstance(unit_ids, list):
                    unit_ids = [unit_ids]
                label = a.get('label', []) # allow this to be a list or an int
                for unit_id in unit_ids:
                    if unit_id not in labels_by_unit:
                        labels_by_unit[unit_id] = []
                    labels_by_unit[unit_id].append(label)
                    labels_by_unit[unit_id] = sorted(list(set(labels_by_unit[unit_id])))
            elif message_type == 'REMOVE_UNIT_LABEL':
                unit_ids = a.get('unitId', '')
                if not isinstance(unit_ids, list):
                    unit_ids = [unit_ids]
                label = a.get('label', '')
                for unit_id in unit_ids:
                    if unit_id in labels_by_unit:
                        labels_by_unit[unit_id] = [x for x in labels_by_unit[unit_id] if x != label]
            elif message_type == 'MERGE_UNITS':
                unit_ids = a.get('unitIds', [])
                merge_groups = _simplify_merge_groups(merge_groups + [unit_ids])
            elif message_type == 'UNMERGE_UNITS':
                unit_ids = a.get('unitIds', [])
                merge_groups = _simplify_merge_groups([[u for u in mg if (u not in unit_ids)] for mg in merge_groups])
            elif message_type == 'CLOSE_CURATION':
                is_closed = True
            elif message_type == 'REOPEN_CURATION':
                is_closed = False
    return {
        'labelsByUnit': labels_by_unit,
        'mergeGroups': merge_groups,
        'isClosed': is_closed
    }
    
def _import_le_recording(subfeed: kc.Subfeed, le_recording):
    messages = _get_messages_from_subfeed(subfeed)
    le_recordings = _get_recordings_from_subfeed_messages(messages)
    id = le_recording['recordingId']
    if id in le_recordings:
        print(f'Recording with ID {id} already exists. Not adding.')
        return
    print(f'Adding recording: {id}')
    subfeed.append_message({
        'action': {
            'type': 'ADD_RECORDING',
            'recording': le_recording
        }
    })

def _import_le_sorting(subfeed: kc.Subfeed, le_sorting):
    messages = _get_messages_from_subfeed(subfeed)
    le_sortings = _get_sortings_from_subfeed_messages(messages)
    id = le_sorting["sortingId"]
    if id in le_sortings:
        print(f'Sorting with ID {id} already exists. Not adding.')
        return
    print(f'Adding sorting: {id}')
    subfeed.append_message({
        'action': {
            'type': 'ADD_SORTING',
            'sorting': le_sorting
        }
    })

def _set_unit_metrics_for_sorting(subfeed: kc.Subfeed, le_unit_metrics_for_sorting):
    sid = le_unit_metrics_for_sorting['sortingId']
    print(f'Setting unit metrics for sorting {sid}')
    subfeed.append_message({
        'action': {
            'type': 'SET_UNIT_METRICS_FOR_SORTING',
            'unitMetricsForSorting': le_unit_metrics_for_sorting
        }
    })

def _delete_recording(*, feed: kc.Feed, recording_id: str):
    subfeed = feed.load_subfeed('main')
    messages = _get_messages_from_subfeed(subfeed)
    le_recordings = _get_recordings_from_subfeed_messages(messages)
    if recording_id not in le_recordings:
        print(f'Cannot remove recording. Recording not found: {recording_id}')
    subfeed.append_message({
        'action': {
            'type': 'DELETE_RECORDINGS',
            'recordingIds': [recording_id]
        }
    })
    le_sortings = _get_sortings_from_subfeed_messages(messages)
    sorting_ids_to_delete = []
    for k, v in le_sortings.items():
        if v.get('recordingId') == recording_id:
            sorting_ids_to_delete.append(v.get('sortingId'))
    if len(sorting_ids_to_delete) > 0:
        subfeed.append_message({
            'action': {
                'type': 'DELETE_SORTINGS',
                'sortingIds': sorting_ids_to_delete
            }
        })

def _query_string_to_dict(q: str):
    ret: Dict[str, str] = {}
    for pstr in q.split('&'):
        vals = pstr.split('=')
        if len(vals) == 2:
            ret[vals[0]] = vals[1]
    return ret

def _dict_to_query_string(x: Dict[str, str]):
    ret = ''
    for k, v in x.items():
        if ret != '':
            ret = ret + '&'
        ret = ret + f'{k}={v}'
    return ret


def _delete_sorting(*, feed: kc.Feed, sorting_id: str):
    subfeed = feed.load_subfeed('main')
    messages = _get_messages_from_subfeed(subfeed)
    le_sortings = _get_recordings_from_subfeed_messages(messages)
    if sorting_id not in le_sortings:
        print(f'Cannot remove sorting. Sorting not found: {sorting_id}')
    subfeed.append_message({
        'action': {
            'type': 'DELETE_SORTINGS',
            'sortingIds': [sorting_id]
        }
    })


from typing import Dict
import numpy as np

def find_unit_neighborhoods(recording, peak_channels_by_unit, max_neighborhood_size) -> Dict[int, list]:
    channel_ids = recording.get_channel_ids()
    locations_by_channel = {}
    for ii, channel_id in enumerate(channel_ids):
        locations_by_channel[int(channel_id)] = np.array(recording.get_channel_property(channel_id=channel_id, property_name='location'))
    neighborhoods = {}
    for unit_id in peak_channels_by_unit.keys():
        if len(channel_ids) <= max_neighborhood_size:
            neighborhood_channel_ids = channel_ids
        else:
            peak_channel_id = peak_channels_by_unit[int(unit_id)]
            peak_location = locations_by_channel[int(peak_channel_id)]
            distances = []
            for channel_id in channel_ids:
                loc = locations_by_channel[int(channel_id)]
                dist = np.linalg.norm(np.array(loc) - np.array(peak_location))
                distances.append(dist)
            sorted_indices = np.argsort(distances)
            neighborhood_channel_ids = []
            for ii in range(min(max_neighborhood_size, len(channel_ids))):
                neighborhood_channel_ids.append(int(channel_ids[sorted_indices[ii]]))
        neighborhoods[int(unit_id)] = neighborhood_channel_ids
    return neighborhoods


import hither2 as hi
from sortingview.config import job_cache, job_handler
from sortingview.helpers import prepare_snippets_h5
import kachery_client as kc

@kc.taskfunction('preload_extract_snippets.2', type='pure-calculation')
def task_preload_extract_snippets(recording_object, sorting_object, snippet_len=(50, 80)):
    with hi.Config(job_handler=job_handler.waveforms, job_cache=job_cache):
        with hi.Config(job_handler=job_handler.extract_snippets):
            snippets_h5 = prepare_snippets_h5.run(recording_object=recording_object, sorting_object=sorting_object, snippet_len=snippet_len)
        return snippets_h5


from typing import Union
from spikeextractors import RecordingExtractor
from spikeextractors import SortingExtractor

import kachery_client as kc
import json
import numpy as np
from .mdaio import DiskReadMda, readmda, readmda_header, writemda32, writemda64, writemda, appendmda
import os


class MdaRecordingExtractor(RecordingExtractor):
    def __init__(self, *, recording_directory=None, timeseries_path: Union[str, None]=None, download=False, samplerate=None, geom=None, geom_path=None, params_path=None):
        RecordingExtractor.__init__(self)
        if recording_directory:
            if timeseries_path is None:
                timeseries_path = recording_directory + '/raw.mda'
            geom_path = recording_directory + '/geom.csv'
            params_path = recording_directory + '/params.json'
        self._timeseries_path = timeseries_path
        if params_path:
            self._dataset_params = kc.load_json(params_path)
            if not self._dataset_params:
                raise Exception('Unable to load recording params: {}'.format(params_path))
            self._samplerate = self._dataset_params['samplerate']
        else:
            self._dataset_params = dict(
                samplerate=samplerate
            )
            self._samplerate = samplerate
        
        if download:
            assert self._timeseries_path is not None
            path0 = kc.load_file(self._timeseries_path)
            if not path0:
                raise Exception('Unable to download file: ' + self._timeseries_path)
        else:
            path0 = self._timeseries_path
        if geom_path is not None:
            if not kc.load_file(geom_path):
                raise Exception('Unable to download file: ' + geom_path)
        if params_path is not None:
            if not kc.load_file(params_path):
                raise Exception('Unable to download file: ' + params_path)
        self._timeseries_path = path0

        self._timeseries = DiskReadMda(self._timeseries_path)
        if self._timeseries is None:
            raise Exception('Unable to load timeseries: {}'.format(self._timeseries_path))
        X = self._timeseries
        if geom is not None:
            self._geom = geom
        elif geom_path:
            geom_path2 = kc.load_file(geom_path)
            self._geom = np.genfromtxt(geom_path2, delimiter=',')
        else:
            self._geom = np.zeros((X.N1(), 2))
        
        if self._geom.shape[0] != X.N1():
            # raise Exception(
            #    'Incompatible dimensions between geom.csv and timeseries file {} <> {}'.format(self._geom.shape[0], X.N1()))
            print('WARNING: Incompatible dimensions between geom.csv and timeseries file {} <> {}'.format(self._geom.shape[0], X.N1()))
            self._geom = np.zeros((X.N1(), 2))
        
        # if self._timeseries_path.startswith('sha1://') or self._timeseries_path.startswith('sha1dir://'):
        #     timeseries_hash_path = self._timeseries_path
        # else:
        #     timeseries_hash_path = 'sha1://' + ka.get_file_hash(self._timeseries_path)
        # self._hash = ka.get_object_hash(dict(
        #     timeseries=timeseries_hash_path,
        #     samplerate=self._samplerate,
        #     geom=_json_serialize(self._geom)
        # ))

        self._num_channels = X.N1()
        self._num_timepoints = X.N2()
        for m in range(self._num_channels):
            self.set_channel_property(m, 'location', self._geom[m, :])
            self.set_channel_property(m, 'group', 0)

    def get_channel_ids(self):
        return list(range(self._num_channels))

    def get_num_frames(self):
        return self._num_timepoints

    def get_sampling_frequency(self):
        return self._samplerate

    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.get_num_frames()
        if channel_ids is None:
            channel_ids = self.get_channel_ids()
        X = self._timeseries
        recordings = X.readChunk(i1=0, i2=start_frame, N1=X.N1(), N2=end_frame - start_frame)
        recordings = recordings[channel_ids, :]
        return recordings
    
    # def hash(self):
    #     return self._hash

    @staticmethod
    def write_recording(recording, save_path, params=dict(), raw_fname='raw.mda', params_fname='params.json', 
            _preserve_dtype=False, in_blocks=False):
        if in_blocks:
            return write_recording_blocks(recording, save_path, params, raw_fname, params_fname, _preserve_dtype)

        channel_ids = recording.get_channel_ids()
        M = len(channel_ids)
        # N = recording.get_num_frames()
        raw = recording.get_traces()
        location0 = recording.get_channel_property(channel_ids[0], 'location')
        nd = len(location0)
        geom = np.zeros((M, nd))
        for ii in range(len(channel_ids)):
            location_ii = recording.get_channel_property(channel_ids[ii], 'location')
            geom[ii, :] = list(location_ii)
        if not os.path.isdir(save_path):
            os.mkdir(save_path)
        if _preserve_dtype:
            writemda(raw, save_path + '/' + raw_fname, dtype=raw.dtype)
        else:
            writemda32(raw, save_path + '/' + raw_fname)
        params["samplerate"] = recording.get_sampling_frequency()
        with open(save_path + '/' + params_fname, 'w') as f:
            json.dump(params, f)
        np.savetxt(save_path + '/geom.csv', geom, delimiter=',')


class MdaSortingExtractor(SortingExtractor):
    def __init__(self, firings_file, samplerate):
        SortingExtractor.__init__(self)
        self._firings_path = kc.load_file(firings_file)
        if not self._firings_path:
            raise Exception('Unable to load firings file: ' + firings_file)

        self._firings = readmda(self._firings_path)
        self._sampling_frequency = samplerate
        self._times = self._firings[1, :]
        self._labels = self._firings[2, :]
        self._unit_ids = np.unique(self._labels).astype(int)

    def get_unit_ids(self):
        return self._unit_ids

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        inds = np.where((self._labels == unit_id) & (start_frame <= self._times) & (self._times < end_frame))
        return np.rint(self._times[inds]).astype(int)

    def get_sampling_frequency(self):
        return self._sampling_frequency

    # def hash(self):
    #     return ka.get_object_hash(dict(
    #         firings=ka.get_file_hash(self._firings_path),
    #         samplerate=self._sampling_frequency
    #     ))
    
    @staticmethod
    def can_read(firings_file: str) -> bool:
        try:
            header = readmda_header(firings_file, verbose=False)
        except:
            return False
        if header is not None and header.num_dims == 2:
            return True
        return False

    @staticmethod
    def write_sorting(sorting, save_path):
        unit_ids = sorting.get_unit_ids()
        # if len(unit_ids) > 0:
        #     K = np.max(unit_ids)
        # else:
        #     K = 0
        times_list = []
        labels_list = []
        for i in range(len(unit_ids)):
            unit = unit_ids[i]
            times = sorting.get_unit_spike_train(unit_id=unit)
            times_list.append(times)
            labels_list.append(np.ones(times.shape) * unit)
        all_times = _concatenate(times_list)
        all_labels = _concatenate(labels_list)
        sort_inds = np.argsort(all_times)
        all_times = all_times[sort_inds]
        all_labels = all_labels[sort_inds]
        L = len(all_times)
        firings = np.zeros((3, L))
        firings[1, :] = all_times
        firings[2, :] = all_labels
        writemda64(firings, save_path)


def _concatenate(list):
    if len(list) == 0:
        return np.array([])
    return np.concatenate(list)


def is_kbucket_url(path):
    path = path or ''
    return path.startswith('kbucket://') or path.startswith('sha1://') or path.startswith('sha1dir://')


def is_url(path):
    path = path or ''
    return path.startswith('http://') or path.startswith('https://') or path.startswith(
        'kbucket://') or path.startswith('sha1://') or path.startswith('sha1dir://')


def write_recording_blocks(recording, save_path, params=dict(), raw_fname='raw.mda', params_fname='params.json',
        _preserve_dtype=False):
    import math
    if not os.path.isdir(save_path):
        os.mkdir(save_path)

    channel_ids = recording.get_channel_ids()
    M = len(channel_ids)
    N = recording.get_num_frames()

    block_size = 20000 * 60 * 10
    n_blocks = math.ceil((N*1.0) / block_size)

    # open file
    total_size = 0
    for i in range(n_blocks):
        i_start = (i*block_size)
        if i == (n_blocks - 1):
            i_end = N
        else:
            i_end = i_start+block_size
        block = recording.get_traces(start_frame=i_start, end_frame=i_end)
        if i == 0:
            writemda32(block, save_path + '/' + raw_fname)
        else:
            appendmda(block, save_path + '/' + raw_fname)
        # write block
        total_size = total_size + np.shape(block)[1]

    location0 = recording.get_channel_property(channel_ids[0], 'location')
    nd = len(location0)
    geom = np.zeros((M, nd))
    for ii in range(len(channel_ids)):
        location_ii = recording.get_channel_property(channel_ids[ii], 'location')
        geom[ii, :] = list(location_ii)

    params["samplerate"] = recording.get_sampling_frequency()
    with open(save_path + '/' + params_fname,'w') as f:
        json.dump(params, f)
    np.savetxt(save_path + '/geom.csv', geom, delimiter=',')

def _listify_ndarray(x):
    if x.ndim == 1:
        if np.issubdtype(x.dtype, np.integer):
            return [int(val) for val in x]
        else:
            return [float(val) for val in x]
    elif x.ndim == 2:
        ret = []
        for j in range(x.shape[1]):
            ret.append(_listify_ndarray(x[:, j]))
        return ret
    elif x.ndim == 3:
        ret = []
        for j in range(x.shape[2]):
            ret.append(_listify_ndarray(x[:, :, j]))
        return ret
    elif x.ndim == 4:
        ret = []
        for j in range(x.shape[3]):
            ret.append(_listify_ndarray(x[:, :, :, j]))
        return ret
    else:
        raise Exception('Cannot listify ndarray with {} dims.'.format(x.ndim))

def _json_serialize(x):
    if isinstance(x, np.ndarray):
        return _listify_ndarray(x)
    elif isinstance(x, np.integer):
        return int(x)
    elif isinstance(x, np.floating):
        return float(x)
    elif type(x) == dict:
        ret = dict()
        for key, val in x.items():
            ret[key] = _json_serialize(val)
        return ret
    elif type(x) == list:
        ret = []
        for i, val in enumerate(x):
            ret.append(_json_serialize(val))
        return ret
    else:
        return x


import seriesview as sev
from ..backend.extensions.spikeamplitudes.spikeamplitudes import runtask_fetch_spike_amplitudes
from ..extractors.labboxephysrecordingextractor import LabboxEphysRecordingExtractor
from ..extractors.labboxephyssortingextractor import LabboxEphysSortingExtractor

class SpikeAmplitudes:
    def __init__(self, *,
        recording: LabboxEphysRecordingExtractor,
        sorting: LabboxEphysSortingExtractor,
        unit_id: int,
        snippet_len=(50, 80)
    ) -> None:
        self._recording = recording
        self._sorting = sorting
        self._unit_id = unit_id
        self._snippet_len = snippet_len
    def prepare_series(self):
        x = runtask_fetch_spike_amplitudes(
            recording=self._recording,
            sorting=self._sorting,
            unit_id=self._unit_id,
            snippet_len=self._snippet_len
        )
        timepoints = x['timepoints']
        amplitudes = x['amplitudes']
        ts = sev.SVSeries.from_numpy(
            type='discrete',
            sampling_frequency=None,
            start_time=0,
            end_time=self._recording.get_num_frames() / self._recording.get_sampling_frequency(),
            segment_duration=10 * 60,
            timestamps=timepoints / self._recording.get_sampling_frequency(),
            values=amplitudes
        )
        return ts


from os import stat
from typing import Any, List, Literal, Union, cast
from .TimeseriesModelDatV1 import TimeseriesModelDatV1
from .TimeseriesModelNwbV1 import TimeseriesModelNwbV1
import numpy as np
import kachery_client as kc
from numpy.core.numeric import Infinity
from .TimeseriesModelNpyV1 import TimeseriesModelNpyV1
from figurl import Figure

class Timeseries:
    def __init__(self, arg: Union[dict, str]):
        if isinstance(arg, str):
            x = kc.load_json(arg)
            if not x:
                raise Exception(f'Unable to load: {arg}')
            arg = cast(dict, x)
        self._model = _load_model(arg)
        self._arg = arg
    @property
    def channel_names(self):
        return self._model.channel_names
    @property
    def channel_properties(self):
        return self._model.channel_properties
    @property
    def num_samples(self):
        return self._model.num_samples
    @property
    def num_channels(self):
        return len(self._model.channel_names)
    @property
    def start_time(self):
        return self._model.start_time
    @property
    def end_time(self):
        return self._model.end_time
    @property
    def type(self) -> Union[Literal['continuous'], Literal['discrete']]:
        return self._model.type
    @property
    def object(self):
        return self._arg
    def get_samples(self, start: Union[None, float]=None, end: Union[None, float]=None, channels: Union[None, List[str]]=None):
        if start is None:
            start = -Infinity
        if end is None:
            end = Infinity
        if channels is None:
            channel_inds = range(self.num_channels)
        else:
            channel_inds = [self.channel_names.index(ch) for ch in channels]
        return self._model.get_samples(start=start, end=end, channel_inds=channel_inds)
    @property
    def sampling_frequency(self):
        return self._model.sampling_frequency
    def figurl(self):
        data = {
            'timeseriesUri': kc.store_json(self.object)
        }
        return Figure(type='experitime.timeseries.1', data=data)
    @staticmethod
    def from_numpy(*, channel_names: List[str], timestamps: np.ndarray, values: np.ndarray, type: Union[Literal['continuous'], Literal['discrete']], channel_properties: Union[dict, None]=None):
        num_channels = len(channel_names)
        assert num_channels == values.shape[1]
        if values.dtype == np.float32:
            values_dtype = 'float32'
        elif values.dtype == np.int16:
            values_dtype = 'int16'
        else:
            raise Exception(f'Unsupported dtype for values: {values.dtype}')
        if timestamps.dtype == np.float32:
            timestamps_dtype = 'float32'
        elif timestamps.dtype == np.float64:
            timestamps_dtype = 'float64'
        else:
            raise Exception(f'Unsupported dtype for timestamps: {values.dtype}')
        with kc.TemporaryDirectory() as tmpdir:
            timestamps_fname = f'{tmpdir}/timestamps.dat'
            timestamps.tofile(timestamps_fname)
            values_fnames: List[str] = []
            for i, channel_name in enumerate(channel_names):
                channel_fname = f'{tmpdir}/channel_{channel_name}.dat'
                values[:, i].tofile(channel_fname)
                values_fnames.append(channel_fname)
            return Timeseries({
                'timeseries_format': 'dat_v1',
                'data': {
                    'channel_names': channel_names,
                    'channel_properties': channel_properties,
                    'num_samples': len(timestamps),
                    'timestamps_uri': kc.store_file(timestamps_fname),
                    'timestamps_dtype': timestamps_dtype,
                    'values_uris': [kc.store_file(values_fnames[i]) for i in range(num_channels)],
                    'values_dtype': values_dtype,
                    'sampling_frequency': _estimate_sampling_frequency(timestamps),
                    'type': type
                }
            })

            # return Timeseries({
            #     'timeseries_format': 'npy_v1',
            #     'data': {
            #         'channel_names': channel_names,
            #         'timestamps_uri': kc.store_npy(timestamps),
            #         'values_uri': kc.store_npy(values),
            #         'type': type
            #     }
            # })
    @staticmethod
    def from_spikeinterface_recording(recording: Any):
        R = recording
        channel_names = [str(id) for id in R.get_channel_ids()]
        timestamps = (np.arange(R.get_num_frames()) * 1 / R.get_sampling_frequency()).astype(np.float32)
        values = R.get_traces().T.astype(np.float32)
        channel_properties = {}
        for id in R.get_channel_ids():
            channel_properties[str(id)] = {'location': np.array(R.get_channel_property(id, 'location')).astype(float).tolist()}
        return Timeseries.from_numpy(
            channel_names=channel_names,
            timestamps=timestamps,
            values=values,
            channel_properties=channel_properties,
            type='continuous'
        )
    @staticmethod
    def from_nwb(nwb_uri: str):
        return Timeseries({
            'timeseries_format': 'nwb_v1',
            'data': {
                'nwb_uri': nwb_uri
            }
        })

def _estimate_sampling_frequency(timestamps: np.ndarray):
    if len(timestamps) <= 1:
        return 0
    deltas = np.diff(timestamps)
    median_delta = np.median(deltas)
    deltas_excluding_outliers = deltas[(median_delta * 0.9 < deltas) & (deltas < median_delta * 1.1)]
    return 1 / float(np.mean(deltas_excluding_outliers))

def _load_model(arg: dict):
    format = arg.get('timeseries_format')
    data = arg.get('data', {})
    if format == 'npy_v1':
        return TimeseriesModelNpyV1(
            channel_names=data['channel_names'],
            timestamps_uri=data['timestamps_uri'],
            values_uri=data['values_uri'],
            type=data['type'],
            sampling_frequency=data['sampling_frequency'],
            channel_properties=data['channel_properties']
        )
    elif format == 'dat_v1':
        return TimeseriesModelDatV1(
            channel_names=data['channel_names'],
            num_samples=data['num_samples'],
            timestamps_uri=data['timestamps_uri'],
            timestamps_dtype=data['timestamps_dtype'],
            values_uris=data['values_uris'],
            values_dtype=data['values_dtype'],
            channel_properties=data['channel_properties'],
            sampling_frequency=data['sampling_frequency'],
            type=data.get('type', 'continuous')
        )
    elif format == 'nwb_v1':
        return TimeseriesModelNwbV1(nwb_uri=data['nwb_uri'])
    else:
        raise Exception(f'Unexpected timeseries format: {format}')



import os
import math
import hither2 as hi
from hither2.dockerimage import RemoteDockerImage
import kachery_client as kc
import numpy as np
from sortingview.config import job_cache, job_handler
from sortingview.serialize_wrapper import serialize_wrapper
from sortingview.helpers import get_unit_waveforms_from_snippets_h5
from sortingview.helpers import prepare_snippets_h5


@kc.taskfunction('individual_cluster_features.1', type='pure-calculation')
def task_individual_cluster_featurest(recording_object, sorting_object, unit_id, snippet_len=(50, 80)):
    with hi.Config(job_handler=job_handler.clusters, job_cache=job_cache):
        with hi.Config(job_handler=job_handler.extract_snippets):
            snippets_h5 = prepare_snippets_h5.run(recording_object=recording_object, sorting_object=sorting_object, snippet_len=snippet_len)
        return individual_cluster_features.run(
            snippets_h5=snippets_h5,
            unit_id=unit_id
        )

@kc.taskfunction('pair_cluster_features.4', type='pure-calculation')
def task_pair_cluster_features(recording_object, sorting_object, unit_id1, unit_id2, snippet_len=(50, 80)):
    with hi.Config(job_handler=job_handler.clusters, job_cache=job_cache):
        with hi.Config(job_handler=job_handler.extract_snippets):
            snippets_h5 = prepare_snippets_h5.run(recording_object=recording_object, sorting_object=sorting_object, snippet_len=snippet_len)
        return pair_cluster_features.run(
            snippets_h5=snippets_h5,
            unit_id1=unit_id1,
            unit_id2=unit_id2
        )


def subsample_inds(n, m):
    if m >= n:
        return range(n)
    incr = n / m
    return [int(math.floor(i * incr)) for i in range(m)]

@hi.function(
    'individual_cluster_features', '0.1.4',
    image=RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
@serialize_wrapper
def individual_cluster_features(snippets_h5, unit_id, max_num_events=1000):
    import h5py
    h5_path = kc.load_file(snippets_h5)
    assert h5_path is not None
    # with h5py.File(h5_path, 'r') as f:
    #     unit_ids = np.array(f.get('unit_ids'))
    #     channel_ids = np.array(f.get('channel_ids'))
    #     channel_locations = np.array(f.get(f'channel_locations'))
    #     sampling_frequency = np.array(f.get('sampling_frequency'))[0].item()
    #     unit_spike_train = np.array(f.get(f'unit_spike_trains/{unit_id}'))
    #     unit_waveforms = np.array(f.get(f'unit_waveforms/{unit_id}/waveforms')) # L x M x T
    #     unit_waveforms_channel_ids = np.array(f.get(f'unit_waveforms/{unit_id}/channel_ids'))
    #     if len(unit_spike_train) > max_num_events:
    #         inds = subsample_inds(len(unit_spike_train), max_num_events)
    #         unit_spike_train = unit_spike_train[inds]
    #         unit_waveforms = unit_waveforms[inds]
    
    unit_waveforms, unit_waveforms_channel_ids, channel_locations0, sampling_frequency, unit_spike_train = get_unit_waveforms_from_snippets_h5(h5_path, unit_id, max_num_events=max_num_events)
    
    from sklearn.decomposition import PCA
    nf = 2 # number of features

    # L = number of waveforms (number of spikes)
    # M = number of electrodes in nbhd
    # T = num. timepoints in the snippet
    W = unit_waveforms # L x M x T

    # subtract mean for each channel and waveform
    for i in range(W.shape[0]):
        for m in range(W.shape[1]):
            W[i, m, :] = W[i, m, :] - np.mean(W[i, m, :])
    X = W.reshape((W.shape[0], W.shape[1] * W.shape[2])) # L x MT
    pca = PCA(n_components=nf)
    pca.fit(X)

    # L = number of waveforms (number of spikes)
    # nf = number of features
    features = pca.transform(X) # L x nf

    return dict(
        timepoints=unit_spike_train.astype(np.float32),
        x=features[:, 0].squeeze().astype(np.float32),
        y=features[:, 1].squeeze().astype(np.float32)
    )

@hi.function(
    'pair_cluster_features', '0.1.4',
    image=RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
@serialize_wrapper
def pair_cluster_features(snippets_h5, unit_id1, unit_id2, max_num_events=1000):
    import h5py
    h5_path = kc.load_file(snippets_h5)
    assert h5_path is not None
    
    # Get the unit waveforms
    unit_waveforms1, unit_waveforms_channel_ids1, channel_locations1, sampling_frequency1, unit_spike_train1 = get_unit_waveforms_from_snippets_h5(h5_path, unit_id1, max_num_events=max_num_events)
    unit_waveforms2, unit_waveforms_channel_ids2, channel_locations2, sampling_frequency2, unit_spike_train2 = get_unit_waveforms_from_snippets_h5(h5_path, unit_id2, max_num_events=max_num_events)

    # Find the channel ids
    channel_ids = [ch for ch in unit_waveforms_channel_ids1 if ch in unit_waveforms_channel_ids2]
    if len(channel_ids) == 0:
        raise Exception('Units do not have any channels in common')
    print(f'Using channel IDs: {channel_ids}')
    
    # Get the indices of the channels
    channel_inds1 = []
    channel_inds2 = []
    for channel_id in channel_ids:
        channel_inds1.append(unit_waveforms_channel_ids1.tolist().index(channel_id))
        channel_inds2.append(unit_waveforms_channel_ids2.tolist().index(channel_id))
    
    # Restrict the unit waveforms to the common channels
    unit_waveforms1 = unit_waveforms1[:, channel_inds1, :] # L1 x M x T
    unit_waveforms2 = unit_waveforms2[:, channel_inds2, :] # L2 x M x T
    L1 = unit_waveforms1.shape[0]
    L2 = unit_waveforms2.shape[0]
    L = L1 + L2

    # Calculate the avg waveforms
    avg_waveform1 = np.mean(unit_waveforms1, axis=0) # M x T
    avg_waveform2 = np.mean(unit_waveforms2, axis=0) # M x T

    # Find the discriminating direction
    discriminating_waveform = avg_waveform2 - avg_waveform1 # M x T
    discriminating_features1 = np.array([np.sum(discriminating_waveform * unit_waveforms1[i, :, :]) for i in range(L1)])
    discriminating_features2 = np.array([np.sum(discriminating_waveform * unit_waveforms2[i, :, :]) for i in range(L2)])
    discriminating_features = np.concatenate((discriminating_features1, discriminating_features2))

    unit_waveforms1b = unit_waveforms1 - avg_waveform1
    unit_waveforms2b = unit_waveforms2 - avg_waveform2

    unit_waveforms_b = np.concatenate((unit_waveforms1b, unit_waveforms2b), axis=0)
    
    assert unit_waveforms_b.shape[0] == L

    unit_spike_train = np.concatenate((unit_spike_train1, unit_spike_train2))

    labels = np.concatenate((np.ones(unit_spike_train1.shape) * unit_id1, np.ones(unit_spike_train2.shape) * unit_id2))
    
    from sklearn.decomposition import PCA
    nf = 1 # number of features

    # L = number of waveforms (number of spikes)
    # M = number of electrodes in nbhd
    # T = num. timepoints in the snippet
    W = unit_waveforms_b # L x M x T

    # subtract mean for each channel and waveform
    for i in range(W.shape[0]):
        for m in range(W.shape[1]):
            W[i, m, :] = W[i, m, :] - np.mean(W[i, m, :])
    X = W.reshape((W.shape[0], W.shape[1] * W.shape[2])) # L x MT
    pca = PCA(n_components=nf)
    pca.fit(X)

    # L = number of waveforms (number of spikes)
    # nf = number of features
    features = pca.transform(X) # L x nf

    x = discriminating_features.astype(np.float32)
    y = features[:, 0].squeeze().astype(np.float32)

    x = x / np.std(x)
    y = y / np.std(y)

    return dict(
        timepoints=unit_spike_train.astype(np.float32),
        x=x,
        y=y,
        labels=labels.astype(np.int32)
    )


import kachery_client as kc
import numpy as np
import spikeextractors as se


class Bin2RecordingExtractor(se.RecordingExtractor):
    extractor_name = 'Bin2RecordingExtractor'
    is_writable = False
    def __init__(self, *, raw, dtype: str, raw_num_channels, num_frames, samplerate, channel_ids, channel_map, channel_positions):
        se.RecordingExtractor.__init__(self)
        
        self._raw = raw
        self._dtype = dtype
        self._num_frames = num_frames
        self._samplerate = samplerate
        self._raw_num_channels = raw_num_channels
        self._channel_ids = channel_ids
        self._channel_map = channel_map
        self._channel_positions = channel_positions
        
        for id in self._channel_ids:
            pos = self._channel_positions[str(id)]
            self.set_channel_property(id, 'location', pos)

    def get_channel_ids(self):
        return self._channel_ids

    def get_num_frames(self):
        return self._num_frames

    def get_sampling_frequency(self):
        return self._samplerate

    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self._num_frames
        if channel_ids is None:
            channel_ids = self._channel_ids
        # M = len(channel_ids)
        # N = end_frame - start_frame

        if self._dtype == 'int16':
            num_bytes_per_entry = 2
        elif self._dtype == 'int32':
            num_bytes_per_entry = 4
        elif self._dtype == 'int64':
            num_bytes_per_entry = 8
        elif self._dtype == 'float32':
            num_bytes_per_entry = 4
        elif self._dtype == 'float64':
            num_bytes_per_entry = 8
        else:
            raise Exception(f'Unexpected dtype: {self._dtype}')
        
        i1 = start_frame * num_bytes_per_entry * self._raw_num_channels
        i2 = end_frame * num_bytes_per_entry * self._raw_num_channels
        
        buf = kc.load_bytes(self._raw, start=i1, end=i2)
        X = np.frombuffer(buf, dtype=self._dtype).reshape((end_frame - start_frame, self._raw_num_channels))
        
        # old method
        # ret = np.zeros((M, N))
        # for ii, ch_id in enumerate(channel_ids):
        #     ret[ii, :] = X[:, self._channel_map[str(ch_id)]]

        # new (equivalent method)
        X = X.T.copy() # this is the part we want to try to speed up
        ret = X[[int(self._channel_map[str(ch_id)]) for ch_id in channel_ids]]
        
        return ret



import kachery_client as kc
import hither2 as hi
from sortingview.config.job_handler import job_handler
from sortingview.config.job_cache import job_cache

@hi.function('fetch_unit_metrics', '0.1.0')
def fetch_unit_metrics(unit_metrics_uri: str):
    return kc.load_json(unit_metrics_uri)

@kc.taskfunction('fetch_unit_metrics.1', type='pure-calculation')
def task_fetch_unit_metrics(unit_metrics_uri: str):
    with hi.Config(job_handler=job_handler.metrics, job_cache=job_cache):
        return fetch_unit_metrics.run(unit_metrics_uri=unit_metrics_uri)



import os
import hither2 as hi
import yaml
from copy import deepcopy

default_config_yaml = '''
job_handlers:
    correlograms:
        type: parallel
        params:
            num_workers: 4
    timeseries:
        type: parallel
        params:
            num_workers: 4
    waveforms:
        type: parallel
        params:
            num_workers: 4
    clusters:
        type: parallel
        params:
            num_workers: 4
    metrics:
        type: parallel
        params:
            num_workers: 4
    misc:
        type: parallel
        params:
            num_workers: 4
    extract_snippets:
        type: parallel
        params:
            num_workers: 4
'''
default_config = yaml.safe_load(default_config_yaml)

config = deepcopy(default_config)

job_handler_config_path = os.getenv('SORTINGVIEW_JOB_HANDLER_CONFIG', None)
if job_handler_config_path is not None:
    print(f'Using job handler config file: {job_handler_config_path}')
    with open(job_handler_config_path, 'r') as f:
        config0 = yaml.safe_load(f)
        config['job_handlers'].update(config0['job_handlers'])
else:
    print('Using default job handler config. To override, set SORTINGVIEW_JOB_HANDLER_CONFIG to path of a yaml file.')

def _job_handler_from_config(x):
    type0 = x['type']
    params0 = x['params']
    if type0 == 'parallel':
        return hi.ParallelJobHandler(params0['num_workers'])
    else:
        raise Exception(f'Invalid type for job handler: {type0}')

print(yaml.safe_dump(config))

class job_handler:
    correlograms = _job_handler_from_config(config['job_handlers']['correlograms'])
    timeseries = _job_handler_from_config(config['job_handlers']['timeseries'])
    waveforms = _job_handler_from_config(config['job_handlers']['waveforms'])
    clusters = _job_handler_from_config(config['job_handlers']['clusters'])
    metrics = _job_handler_from_config(config['job_handlers']['metrics'])
    misc = _job_handler_from_config(config['job_handlers']['misc'])
    extract_snippets = _job_handler_from_config(config['job_handlers']['extract_snippets'])


from abc import ABC, abstractmethod
import spikeextractors as se
import numpy as np


class FilterRecording(se.RecordingExtractor):
    def __init__(self, *, recording, chunk_size=10000):
        se.RecordingExtractor.__init__(self)
        self._recording = recording
        self._chunk_size = chunk_size
        self.copy_channel_properties(recording)

    def paramsForHash(self):
        return None

    def get_channel_ids(self):
        return self._recording.get_channel_ids()

    def get_num_frames(self):
        return self._recording.get_num_frames()

    def get_sampling_frequency(self):
        return self._recording.get_sampling_frequency()

    def get_traces(self, start_frame=None, end_frame=None, channel_ids=None, return_scaled=True):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.get_num_frames()
        if channel_ids is None:
            channel_ids = self.get_channel_ids()
        ich1 = int(start_frame / self._chunk_size)
        ich2 = int((end_frame - 1) / self._chunk_size)
        filtered_chunk_list = []
        for ich in range(ich1, ich2 + 1):
            filtered_chunk0 = self._get_filtered_chunk(ich)
            if ich == ich1:
                start0 = start_frame - ich * self._chunk_size
            else:
                start0 = 0
            if ich == ich2:
                end0 = end_frame - ich * self._chunk_size
            else:
                end0 = self._chunk_size
            chan_idx = [self.get_channel_ids().index(chan) for chan in channel_ids]
            filtered_chunk_list.append(filtered_chunk0[chan_idx, start0:end0])
        if len(filtered_chunk_list) > 0:
            return np.concatenate(filtered_chunk_list, axis=1)
        else:
            return np.zeros((len(channel_ids), 0))

    @abstractmethod
    def filterChunk(self, *, start_frame, end_frame):
        raise NotImplementedError('filterChunk not implemented')

    def _get_filtered_chunk(self, ind):
        start0 = ind * self._chunk_size
        end0 = (ind + 1) * self._chunk_size
        chunk1 = self.filterChunk(start_frame=start0, end_frame=end0)
        return chunk1



import os
import json
from typing import Union
import kachery_client as kc

thisdir = os.path.dirname(os.path.realpath(__file__))
with open(thisdir + '/task_function_ids.json', 'r') as f:
    task_function_ids = json.load(f)

def start_backend(*, channel: str, backend_id: Union[str, None]=None):
    # register the tasks
    from ..tasks import dummy
    from .extensions import dummy

    kc.run_task_backend(
        channels=[channel],
        task_function_ids=task_function_ids,
        backend_id=backend_id
    )


from .version import __version__
from .backend.start_backend_cli import start_backend_cli
from .backend.start_backend import start_backend
from .workspace_list import WorkspaceList
from .user_permissions import set_user_permissions, set_user_feed_permissions, get_user_permissions_dict
from .workspace import load_workspace, Workspace
from .workspace.workspace import create_workspace
from .serialize_wrapper import serialize_wrapper
from .extractors import LabboxEphysRecordingExtractor, LabboxEphysSortingExtractor
from .workspace_list import get_workspace_list, set_workspace_list, add_workspace_to_list
from .extractors import H5SortingExtractorV1
from .extractors.h5extractors.h5recordingextractorv1 import H5RecordingExtractorV1
from .extractors.subrecording import subrecording
from .extractors.subsorting import subsorting
from .extractors.wrapperrecordingextractor import create_recording_from_old_extractor
# from .SpikeAmplitudes import SpikeAmplitudes
# from .SpikeExplorer import SpikeExplorer


from time import time
from typing import List, Literal, Union
import kachery_client as kc
import numpy as np
from ._estimate_sampling_frequency import _estimate_sampling_frequency

class TimeseriesModelNwbV1:
    def __init__(self, *, nwb_uri: str, electrical_series_name: Union[str, None]=None):
        self._path = kc.load_file(nwb_uri)
        assert self._path is not None, f'Unable to load nwb file: {nwb_uri}'
        self._nwb_uri = nwb_uri
        self._electrical_series_name = electrical_series_name

        self._initialize()
    def _initialize(self):
        from pynwb import NWBHDF5IO
        from pynwb.ecephys import ElectrodeGroup
        with NWBHDF5IO(self._path, 'r') as io:
            nwbfile = io.read()
            if self._electrical_series_name is None:
                a_names = list(nwbfile.acquisition)
                if len(a_names) > 1:
                    raise ValueError('More than one acquisition found. You must specify electrical_series.')
                if len(a_names) == 0:
                    raise ValueError('No acquisitions found in the .nwb file.')
                self._electrical_series_name = a_names[0]
            es = nwbfile.acquisition[self._electrical_series_name]
            if hasattr(es, 'timestamps') and es.timestamps:
                self._timestamps: np.ndarray = es.timestamps
                self._sampling_frequency = _estimate_sampling_frequency(self._timestamps)
            else:
                self._sampling_frequency = es.rate
                if hasattr(es, 'starting_time'):
                    start_time = es.starting_time
                else:
                    start_time = 0.
                num_samples = es.data.shape[0]
                self._timestamps = start_time + np.arange(num_samples) / self._sampling_frequency
            
            num_channels = len(es.electrodes.table.id[:])

            # Channels gains - for RecordingExtractor, these are values to cast traces to uV
            if es.channel_conversion is not None:
                gains = es.conversion * es.channel_conversion[:] * 1e6
            else:
                gains = es.conversion * np.ones(num_channels) * 1e6

            # # Extractors channel groups must be integers, but Nwb electrodes group_name can be strings
            # if 'group_name' in nwbfile.electrodes.colnames:
            #     unique_grp_names = list(np.unique(nwbfile.electrodes['group_name'][:]))

            # Fill channel properties dictionary from electrodes table
            self._channel_names = [str(c) for c in es.electrodes.table.id[es.electrodes.data]]
            self._channel_properties = {}
            for c in self._channel_names:
                self._channel_properties[c] = {}
            for es_ind, (channel_name, electrode_table_index) in enumerate(zip(self._channel_names, es.electrodes.data)):
                self._channel_properties[channel_name]['gain'] = gains[es_ind]
                this_loc = []
                if 'rel_x' in nwbfile.electrodes:
                    this_loc.append(nwbfile.electrodes['rel_x'][electrode_table_index])
                    if 'rel_y' in nwbfile.electrodes:
                        this_loc.append(nwbfile.electrodes['rel_y'][electrode_table_index])
                    else:
                        this_loc.append(0)
                    self._channel_properties[channel_name]['location'] = this_loc

                for col in nwbfile.electrodes.colnames:
                    if isinstance(nwbfile.electrodes[col][electrode_table_index], ElectrodeGroup):
                        continue
                    elif col == 'group_name':
                        self._channel_properties[channel_name]['group_name'] = nwbfile.electrodes[col][electrode_table_index]
                    elif col == 'location':
                        self._channel_properties[channel_name]['brain_area'] = nwbfile.electrodes[col][electrode_table_index]
                    elif col in ['x', 'y', 'z', 'rel_x', 'rel_y', 'group_name', 'location']:
                        continue
                    else:
                        self._channel_properties[channel_name][col] = nwbfile.electrodes[col][electrode_table_index]

            # Fill epochs dictionary
            self._epochs = {}
            if nwbfile.epochs is not None:
                df_epochs = nwbfile.epochs.to_dataframe()
                self._epochs = {row['tags'][0]: {
                    'start_frame': row['start_time'],
                    'end_frame': row['stop_time']}
                    for _, row in df_epochs.iterrows()}

            
            # Make metadata
            # Metadata dictionary - useful for constructing a nwb file
            self._nwb_metadata = dict()
            self._nwb_metadata['NWBFile'] = {
                'session_description': nwbfile.session_description,
                'identifier': nwbfile.identifier,
                'session_start_time': nwbfile.session_start_time,
                'institution': nwbfile.institution,
                'lab': nwbfile.lab
            }
            self._nwb_metadata['Ecephys'] = dict()
            # Update metadata with Device info
            self._nwb_metadata['Ecephys']['Device'] = []
            for dev in nwbfile.devices:
                self._nwb_metadata['Ecephys']['Device'].append({'name': dev})
            # Update metadata with ElectrodeGroup info
            self._nwb_metadata['Ecephys']['ElectrodeGroup'] = []
            for k, v in nwbfile.electrode_groups.items():
                self._nwb_metadata['Ecephys']['ElectrodeGroup'].append({
                    'name': v.name,
                    'description': v.description,
                    'location': v.location,
                    'device': v.device.name
                })
            # Update metadata with ElectricalSeries info
            self._nwb_metadata['Ecephys']['ElectricalSeries'] = []
            self._nwb_metadata['Ecephys']['ElectricalSeries'].append({
                'name': es.name,
                'description': es.description
            })
    @property
    def channel_names(self):
        return self._channel_names
    @property
    def channel_properties(self):
        return self._channel_properties
    @property
    def num_samples(self):
        return len(self._timestamps)
    @property
    def start_time(self):
        return self._timestamps[0]
    @property
    def end_time(self):
        return self._timestamps[-1]
    @property
    def type(self) -> Union[Literal['continuous'], Literal['discrete']]:
        return 'continuous'
    def get_samples(self, start: float, end: float, channel_inds: Union[List[int], range]):
        from pynwb import NWBHDF5IO
        with NWBHDF5IO(self._path, 'r') as io:
            nwbfile = io.read()
            es = nwbfile.acquisition[self._electrical_series_name]

            traces = np.array([])
            # fix this part
            # es_channel_ids = np.array(es.electrodes.table.id[:])[es.electrodes.data[:]].tolist()
            # channel_inds = [es_channel_ids.index(id) for id in channel_ids]
            # if np.array(channel_ids).size > 1 and np.any(np.diff(channel_ids) < 0):
            #     # get around h5py constraint that it does not allow datasets
            #     # to be indexed out of order
            #     sorted_channel_ids = np.sort(channel_ids)
            #     sorted_idx = np.array([list(sorted_channel_ids).index(ch) for ch in channel_ids])
            #     recordings = es.data[start_frame:end_frame, sorted_channel_ids].T
            #     traces = recordings[sorted_idx, :]
            # else:
            #     traces = es.data[start_frame:end_frame, channel_inds].T
            # # This DatasetView and lazy operations will only work within context
            # # We're keeping the non-lazy version for now
            # # es_view = DatasetView(es.data)  # es is an instantiated h5py dataset
            # # traces = es_view.lazy_slice[start_frame:end_frame, channel_ids].lazy_transpose()
        return traces
    @property
    def sampling_frequency(self):
        return self._sampling_frequency



import kachery_client as kc
from sortingview.serialize_wrapper import serialize_wrapper

@kc.taskfunction('sortingview.get_snippets_for_uri.1', type='pure-calculation')
@serialize_wrapper
def task_get_snippets_for_uri(snippets_uri: str):
    return kc.load_npy(snippets_uri)

@kc.taskfunction('sortingview.get_timestamps_for_uri.1', type='pure-calculation')
@serialize_wrapper
def task_get_timestamps_for_uri(timestamps_uri: str):
    return kc.load_npy(timestamps_uri)

@kc.taskfunction('sortingview.get_features_for_uri.1', type='pure-calculation')
@serialize_wrapper
def task_get_features_for_uri(features_uri: str):
    return kc.load_npy(features_uri)


from time import time
from typing import Dict, List, Literal, Union
import kachery_client as kc
import numpy as np
from numpy.core.numeric import Infinity

class TimeseriesModelDatV1:
    def __init__(self, *, channel_names: List[str], num_samples: int, timestamps_uri: str, timestamps_dtype: str, values_uris: List[str], values_dtype: str, channel_properties: Union[Dict[str, dict], None]=None, sampling_frequency: float, type: Union[Literal['continuous'], Literal['discrete']]):
        self._channel_names = channel_names
        self._channel_properties = channel_properties
        self._timestamps = _load_timestamps_from_uri(timestamps_uri, dtype=timestamps_dtype)
        assert len(self._timestamps) == num_samples, 'Mismatch between num_samples and length of timestamps'
        self._num_channels = len(channel_names)
        self._values: List[np.ndarray] = [
            _load_values_from_uri(values_uris[i], dtype=values_dtype)
            for i in range(self._num_channels)
        ]
        self._type: Union[Literal['continuous'], Literal['discrete']] = type
        self._sampling_frequency = sampling_frequency
    @property
    def channel_names(self):
        return self._channel_names
    @property
    def channel_properties(self):
        return self._channel_properties
    @property
    def num_samples(self):
        return len(self._timestamps)
    @property
    def start_time(self):
        return self._timestamps[0]
    @property
    def end_time(self):
        return self._timestamps[-1]
    @property
    def type(self) -> Union[Literal['continuous'], Literal['discrete']]:
        return self._type
    def get_samples(self, start: float, end: float, channel_inds: Union[List[int], range]):
        x = np.where((start <= self._timestamps) & (self._timestamps < end))[0]
        ret = np.zeros((len(x), len(channel_inds)), dtype=self._values[0].dtype)
        print(self._values[0].shape, self._timestamps.shape)
        for i, channel_ind in enumerate(channel_inds):
            ret[:, i] = self._values[channel_ind][x]
        return self._timestamps[x], ret
    @property
    def sampling_frequency(self):
        return self._sampling_frequency

def _load_timestamps_from_uri(uri: str, *, dtype: str):
    if dtype == 'float32':
        d = np.float32
    elif dtype == 'float64':
        d = np.float64
    else:
        raise Exception(f'Unexpected dtype for loading timestamps: {dtype}')
    local_fname = kc.load_file(uri)
    if local_fname is None:
        raise Exception(f'Unable to load timestamps: {uri}')
    return np.fromfile(local_fname, dtype=d)

def _load_values_from_uri(uri: str, *, dtype: str):
    if dtype == 'float32':
        d = np.float32
    elif dtype == 'int16':
        d = np.int16
    else:
        raise Exception(f'Unexpected dtype for loading values: {dtype}')
    local_fname = kc.load_file(uri)
    if local_fname is None:
        raise Exception(f'Unable to load values: {uri}')
    return np.fromfile(local_fname, dtype=d)


from typing import Any
import kachery_client as kc
from ..version import __version__

@kc.taskfunction('sortingview.get_python_package_version.1', type='query')
def task_sortingview_get_python_package_version():
    return __version__


import os
import hither2 as hi
import kachery_client as kc
import numpy as np
from sortingview.serialize_wrapper import serialize_wrapper
from sortingview.helpers import get_unit_waveforms_from_snippets_h5


@hi.function(
    'get_sorting_unit_snippets', '0.1.8',
    image=hi.RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
@serialize_wrapper
def get_sorting_unit_snippets(snippets_h5, unit_id, time_range, max_num_snippets):
    import h5py
    h5_path = kc.load_file(snippets_h5)
    assert h5_path is not None
    # with h5py.File(h5_path, 'r') as f:
    #     unit_ids = np.array(f.get('unit_ids'))
    #     channel_ids = np.array(f.get('channel_ids'))
    #     channel_locations = np.array(f.get(f'channel_locations'))
    #     sampling_frequency = np.array(f.get('sampling_frequency'))[0].item()
    #     if np.isnan(sampling_frequency):
    #         print('WARNING: sampling frequency is nan. Using 30000 for now. Please correct the snippets file.')
    #         sampling_frequency = 30000
    #     unit_spike_train = np.array(f.get(f'unit_spike_trains/{unit_id}'))
    #     unit_waveforms = np.array(f.get(f'unit_waveforms/{unit_id}/waveforms'))
    #     unit_waveforms_channel_ids = np.array(f.get(f'unit_waveforms/{unit_id}/channel_ids'))
    #     print(unit_waveforms_channel_ids)
    unit_waveforms, unit_waveforms_channel_ids, channel_locations0, sampling_frequency, unit_spike_train = get_unit_waveforms_from_snippets_h5(h5_path, unit_id)
    
    snippets = [
        {
            'index': j,
            'unitId': unit_id,
            'waveform': unit_waveforms[j].astype(np.float32),
            'timepoint': float(unit_spike_train[j])
        }
        for j in range(unit_waveforms.shape[0])
        if time_range['min'] <= unit_spike_train[j] and unit_spike_train[j] < time_range['max']
    ]

    return dict(
        channel_ids=unit_waveforms_channel_ids.astype(np.int32),
        channel_locations=channel_locations0.astype(np.float32),
        sampling_frequency=sampling_frequency,
        snippets=snippets[:max_num_snippets]
    )

@hi.function(
    'get_sorting_unit_info', '0.1.1',
    image=hi.RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
@serialize_wrapper
def get_sorting_unit_info(snippets_h5, unit_id):
    import h5py
    h5_path = kc.load_file(snippets_h5)
    assert h5_path is not None
    # with h5py.File(h5_path, 'r') as f:
    #     unit_ids = np.array(f.get('unit_ids'))
    #     channel_ids = np.array(f.get('channel_ids'))
    #     channel_locations = np.array(f.get(f'channel_locations'))
    #     sampling_frequency = np.array(f.get('sampling_frequency'))[0].item()
    #     if np.isnan(sampling_frequency):
    #         print('WARNING: sampling frequency is nan. Using 30000 for now. Please correct the snippets file.')
    #         sampling_frequency = 30000
    #     unit_waveforms_channel_ids = np.array(f.get(f'unit_waveforms/{unit_id}/channel_ids'))
    #     print(unit_waveforms_channel_ids)
    unit_waveforms, unit_waveforms_channel_ids, channel_locations0, sampling_frequency, unit_spike_train = get_unit_waveforms_from_snippets_h5(h5_path, unit_id)
    
    channel_locations_2 = []
    for ch_id in unit_waveforms_channel_ids:
        ind = np.where(unit_waveforms_channel_ids == ch_id)[0]
        channel_locations_2.append(channel_locations0[ind].ravel().tolist())

    return dict(
        channel_ids=unit_waveforms_channel_ids.astype(np.int32),
        channel_locations=channel_locations_2,
        sampling_frequency=sampling_frequency
    )



dummy = 0
from .example_recording_sorting import *
from .recording_info import *
from .sorting_info import *
from .preload_extract_snippets import *
from .workspace_list_subfeed import *
from .workspace_action import *
from .workspace_sorting_curation_action import *
from .latency_test import *
from .python_package_version import *
from .get_workspace_list import *
from .get_figure_object import *
from .get_experitime_timeseries_samples import *
from .get_snippets_for_uri import *